import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="501",
    message="A new episode has started before the end date of the previous episode.",
    affected_fields=["DECOM", "DEC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi = epi.reset_index()
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi["DEC"] = pd.to_datetime(epi["DEC"], format="%d/%m/%Y", errors="coerce")

        epi = epi.sort_values(["CHILD", "DECOM"])

        epi_lead = epi.shift(1)
        epi_lead = epi_lead.reset_index()

        m_epi = epi.merge(
            epi_lead, left_on="index", right_on="level_0", suffixes=("", "_prev")
        )

        error_cohort = m_epi[
            (m_epi["CHILD"] == m_epi["CHILD_prev"])
            & (m_epi["DECOM"] < m_epi["DEC_prev"])
        ]
        error_list = error_cohort["index"].to_list()
        error_list.sort()
        return {"Episodes": error_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "111", "DECOM": "01/06/2020", "DEC": "04/06/2020"},  # 0
            {"CHILD": "111", "DECOM": "02/06/2020", "DEC": "06/06/2020"},  # 1   Fails
            {"CHILD": "111", "DECOM": "06/06/2020", "DEC": pd.NA},  # 2
            {"CHILD": "111", "DECOM": "08/06/2020", "DEC": "09/06/2020"},  # 3
            {"CHILD": "222", "DECOM": "10/06/2020", "DEC": "11/06/2020"},  # 4
            {"CHILD": "333", "DECOM": "04/06/2020", "DEC": "07/06/2020"},  # 5
            {"CHILD": "333", "DECOM": "05/06/2020", "DEC": pd.NA},  # 6   Fails
            {"CHILD": "444", "DECOM": "08/06/2020", "DEC": "09/06/2020"},  # 7
            {"CHILD": "444", "DECOM": "08/06/2020", "DEC": "10/06/2020"},  # 8   Fails
            {"CHILD": "444", "DECOM": "15/06/2020", "DEC": pd.NA},  # 9
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 6, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="563",
    message="The child should no longer be placed for adoption but the date of the decision that the child should be placed for adoption is blank",
    affected_fields=["DATE_PLACED", "REASON_PLACED_CEASED", "DATE_PLACED_CEASED"],
    tables=["PlacedAdoption"],
)
def validate(dfs):
    if "PlacedAdoption" not in dfs:
        return {}
    else:
        placed_adoption = dfs["PlacedAdoption"]
        mask = (
            placed_adoption["REASON_PLACED_CEASED"].notna()
            & placed_adoption["DATE_PLACED_CEASED"].notna()
            & placed_adoption["DATE_PLACED"].isna()
        )
        error_locations = placed_adoption.index[mask]
        return {"PlacedAdoption": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_placed = pd.DataFrame(
        [
            {
                "CHILD": "11",
                "DATE_PLACED_CEASED": pd.NA,
                "REASON_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "26/05/2000",
            },  # 0
            {
                "CHILD": "202",
                "DATE_PLACED_CEASED": "01/02/2003",
                "REASON_PLACED_CEASED": "invalid dont matter",
                "DATE_PLACED": "26/05/2000",
            },  # 1
            {
                "CHILD": "3003",
                "DATE_PLACED_CEASED": pd.NA,
                "REASON_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "26/05/2000",
            },  # 2
            {
                "CHILD": "40004",
                "DATE_PLACED_CEASED": "26/05/2001",
                "REASON_PLACED_CEASED": "some sample reason",
                "DATE_PLACED": pd.NA,
            },  # 3: Fail
            {
                "CHILD": "606",
                "DATE_PLACED_CEASED": pd.NA,
                "REASON_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "26/05/2000",
            },  # 4
        ]
    )

    fake_dfs = {"PlacedAdoption": fake_placed}

    result = validate(fake_dfs)
    assert result == {"PlacedAdoption": [3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="555",
    message="Freeing order has been granted but there is no date of decision that the child should "
    + "be placed for adoption.",
    affected_fields=[
        "CHILD",
        "DATE_PLACED",
        "DATE_PLACED_CEASED",
        "REASON_PLACED_CEASED",
    ],
    tables=["Episodes", "PlacedAdoption"],
)
def validate(dfs):
    if "Episodes" not in dfs or "PlacedAdoption" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        sho = dfs["PlacedAdoption"]
        sho.reset_index(inplace=True)
        epi.reset_index(inplace=True)

        # D1 episodes without a corresponding PlacedAdoption entry
        epi_has_d1 = epi[epi["LS"] == "D1"]
        merge_w_sho = epi_has_d1.merge(
            sho, how="left", on="CHILD", suffixes=["_EP", "_PA"], indicator=True
        )
        err_list_epi = (
            merge_w_sho.query("_merge == 'left_only'")["index_EP"].unique().tolist()
        )

        # Open D1 Episodes where DATE_PLACED_CEASED or REASON_PLACED_CEASED is filled in
        epi_open_d1 = epi[(epi["LS"] == "D1") & epi["DEC"].isna()]
        merge_w_sho2 = epi_open_d1.merge(
            sho, how="inner", on="CHILD", suffixes=["_EP", "_PA"]
        )
        err_list_sho = merge_w_sho2["index_PA"][
            merge_w_sho2["DATE_PLACED_CEASED"].notna()
            | merge_w_sho2["REASON_PLACED_CEASED"].notna()
        ]
        err_list_sho = err_list_sho.unique().tolist()
        return {"Episodes": err_list_epi, "PlacedAdoption": err_list_sho}


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        [
            {"CHILD": "111", "LS": "E1", "DEC": "01/06/2020", "PLACE": "T3"},  # 0
            {"CHILD": "222", "LS": "D1", "DEC": "05/06/2020", "PLACE": "P3"},  # 1
            {"CHILD": "333", "LS": "D1", "DEC": "06/06/2020", "PLACE": "T3"},  # 2
            {
                "CHILD": "444",
                "LS": "D1",
                "DEC": pd.NA,
                "PLACE": "T3",
            },  # 3 Err not in sho
            {
                "CHILD": "555",
                "LS": "D1",
                "DEC": pd.NA,
                "PLACE": "T3",
            },  # 4 Err not in sho
            {"CHILD": "666", "LS": "D1", "DEC": pd.NA, "PLACE": "T3"},  # 5
        ]
    )
    fake_sho = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DATE_PLACED": "01/06/2020",
                "DATE_PLACED_CEASED": "01/06/2020",  # 0
                "REASON_PLACED_CEASED": "A",
            },
            {
                "CHILD": "222",
                "DATE_PLACED": "01/06/2020",
                "DATE_PLACED_CEASED": "01/06/2020",  # 1
                "REASON_PLACED_CEASED": "A",
            },
            {
                "CHILD": "333",
                "DATE_PLACED": "01/06/2020",
                "DATE_PLACED_CEASED": "01/06/2020",  # 2
                "REASON_PLACED_CEASED": "A",
            },
            {
                "CHILD": "666",
                "DATE_PLACED": "01/06/2020",
                "DATE_PLACED_CEASED": pd.NA,  # 3 err ceased not null
                "REASON_PLACED_CEASED": "A",
            },
            {
                "CHILD": "123",
                "DATE_PLACED": "01/06/2020",
                "DATE_PLACED_CEASED": "01/06/2020",  # 4
                "REASON_PLACED_CEASED": "A",
            },
            {
                "CHILD": "984",
                "DATE_PLACED": pd.NA,
                "DATE_PLACED_CEASED": "01/06/2020",  # 5
                "REASON_PLACED_CEASED": "A",
            },
        ]
    )
    fake_dfs = {"Episodes": fake_epi, "PlacedAdoption": fake_sho}

    result = validate(fake_dfs)

    assert result == {"Episodes": [3, 4], "PlacedAdoption": [3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="101",
    message="Gender code is not valid.",
    affected_fields=["SEX"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}

    header = dfs["Header"]
    code_list = ["1", "2"]

    mask = header["SEX"].astype(str).isin(code_list)

    validation_error_mask = ~mask
    validation_error_locations = header.index[validation_error_mask]

    return {"Header": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SEX": [1, 2, 3, pd.NA],
        }
    )

    fake_dfs = {"Header": fake_data}

    result = validate(fake_dfs)

    assert result == {"Header": [2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1005",
    message="The end date of the missing episode or episode that the child was away from placement without authorisation is not a valid date.",
    affected_fields=["MIS_END"],
    tables=["Missing"],
)
def validate(dfs):
    if "Missing" not in dfs:
        return {}
    else:
        missing = dfs["Missing"]
        missing["fMIS_END"] = pd.to_datetime(
            missing["MIS_END"], format="%d/%m/%Y", errors="coerce"
        )

        missing_end_date = missing["MIS_END"].isna()
        invalid_end_date = missing["fMIS_END"].isna()

        error_mask = ~missing_end_date & invalid_end_date

        error_locations = missing.index[error_mask]

        return {"Missing": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "MIS_START": [
                "08/03/2020",
                "22/06/2020",
                pd.NA,
                "13/10/2021",
                "10/24/2021",
            ],
            "MIS_END": ["08/03/2020", pd.NA, "22/06/2020", "13/10/21", pd.NA],
        }
    )

    fake_dfs = {"Missing": fake_data}

    result = validate(fake_dfs)

    assert result == {"Missing": [3]}



"""
Column names for each 903 file.

These are used in two places:
- Checking that the uploaded files match and identifying which CSV file is which (when CSVs are selected).
- Constructing CSV files from provided XML (when XML is selected)
"""

column_names = {
    "Header": ["CHILD", "SEX", "DOB", "ETHNIC", "UPN", "MOTHER", "MC_DOB"],
    "Episodes": [
        "CHILD",
        "DECOM",
        "RNE",
        "LS",
        "CIN",
        "PLACE",
        "PLACE_PROVIDER",
        "DEC",
        "REC",
        "REASON_PLACE_CHANGE",
        "HOME_POST",
        "PL_POST",
        "URN",
    ],
    "Reviews": ["CHILD", "DOB", "REVIEW", "REVIEW_CODE"],
    "UASC": ["CHILD", "SEX", "DOB", "DUC"],
    "OC2": [
        "CHILD",
        "DOB",
        "SDQ_SCORE",
        "SDQ_REASON",
        "CONVICTED",
        "HEALTH_CHECK",
        "IMMUNISATIONS",
        "TEETH_CHECK",
        "HEALTH_ASSESSMENT",
        "SUBSTANCE_MISUSE",
        "INTERVENTION_RECEIVED",
        "INTERVENTION_OFFERED",
    ],
    "OC3": ["CHILD", "DOB", "IN_TOUCH", "ACTIV", "ACCOM"],
    "AD1": [
        "CHILD",
        "DOB",
        "DATE_INT",
        "DATE_MATCH",
        "FOSTER_CARE",
        "NB_ADOPTR",
        "SEX_ADOPTR",
        "LS_ADOPTR",
    ],
    "PlacedAdoption": [
        "CHILD",
        "DOB",
        "DATE_PLACED",
        "DATE_PLACED_CEASED",
        "REASON_PLACED_CEASED",
    ],
    # Note: for PrevPerm, LA_PERM will usually be blank and shouldn't be used directly
    "PrevPerm": ["CHILD", "DOB", "PREV_PERM", "LA_PERM", "DATE_PERM"],
    "Missing": ["CHILD", "DOB", "MISSING", "MIS_START", "MIS_END"],
    "SWEpisodes": ["CHILD", "DOB", "SW_ID", "SW_DECOM", "SW_DEC", "SW_REASON"],
}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW13STG1",
    message="The reason for social worker change is either not valid or has not been entered.",
    affected_fields=["SW_REASON"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]

        valid_reasons = [
            "MANAGE",
            "FCONTA",
            "LEFTRL",
            "ORGRST",
            "TSPROC",
            "ABSENC",
            "CHCHAN",
            "PCCHAN",
            "SWDIED",
            "OTHERS",
        ]

        error_rows = df[~df["SW_REASON"].isin(valid_reasons)].index

        return {"SWEpisodes": error_rows.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SW_REASON": ["LAZINESS", "SWDIED", "SLEEPINESS", "CHCHAN"],
        }
    )

    fake_dfs = {"SWEpisodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [0, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="406",
    message="Child is Unaccompanied Asylum-Seeking Child (UASC) or was formerly UASC. Distance should be "
    "blank. [NOTE: This check will result in false negatives for children formerly UASC not "
    "identified as such in loaded data]",
    affected_fields=["PL_DISTANCE"],
    tables=["Episodes"],
)

# If <UASC> = '1' or <COLLECTION YEAR> - 1 <UASC> = '1' or <COLLECTION YEAR> - 2 <UASC> = '1' Then
# <PL_DISTANCE> should not be provided Note: <PL_DISTANCE> is a derived field in most instances
def validate(dfs):
    # Header_last also used if present
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]

        # Note the initial positions. Freeze a copy of the index values into a column
        epi["orig_idx"] = epi.index

        # Work out who is formerly
        header = pd.DataFrame()
        if "Header" in dfs:
            header_current = dfs["Header"]
            header = pd.concat((header, header_current), axis=0)
        elif "UASC" in dfs:
            uasc = dfs["UASC"]
            uasc = uasc.loc[
                uasc.drop("CHILD", axis="columns").notna().any(axis=1), ["CHILD"]
            ].copy()
            uasc.loc[:, "UASC"] = "1"
            header = pd.concat((header, uasc), axis=0)

        if "Header_last" in dfs:
            header = pd.concat((header, dfs["Header_last"]), axis=0)
        elif "UASC_last" in dfs:
            uasc = dfs["UASC_last"]
            uasc = uasc.loc[
                uasc.drop("CHILD", axis="columns").notna().any(axis=1), ["CHILD"]
            ].copy()
            uasc.loc[:, "UASC"] = "1"
            header = pd.concat((header, uasc), axis=0)

        if "UASC" in header.columns:
            header = header[header.UASC == "1"].drop_duplicates("CHILD")
            epi = epi.merge(header[["CHILD"]], how="inner", on="CHILD")
        else:
            # if theres no UASC column in header, either from XML data, inferred for CSV in ingress, or added above
            # then we can't identify anyone as UASC/formerly UASC
            return {}
        # PL_DISTANCE is added when the uploaded files are read into the tool. The code that does this is found in datastore.py

        err_list = (
            epi.loc[epi["PL_DISTANCE"].notna(), "orig_idx"]
            .sort_values()
            .unique()
            .tolist()
        )

        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_episodes = pd.DataFrame(
        [
            {"CHILD": "111", "PL_DISTANCE": "XX1"},  # 0 fail
            {"CHILD": "222", "PL_DISTANCE": pd.NA},  # 1
            {"CHILD": "222", "PL_DISTANCE": "XX1"},  # 2 fail
            {"CHILD": "333", "PL_DISTANCE": pd.NA},  # 3
            {
                "CHILD": "333",
                "PL_DISTANCE": "XX1",
            },  # 4 ignore: it is not present in any uasc table.
            {"CHILD": "345", "PL_DISTANCE": pd.NA},  # 5
            {"CHILD": "444", "PL_DISTANCE": "XX1"},  # 6 fail
            {"CHILD": "444", "PL_DISTANCE": pd.NA},  # 7
        ]
    )
    fake_header = pd.DataFrame(
        [
            {"CHILD": "111", "UASC": "1"},  # 0
            {"CHILD": "222", "UASC": "0"},  # 2
            {"CHILD": "333", "UASC": "0"},  # 4
            {"CHILD": "345", "UASC": "1"},  # 5
            {"CHILD": "444", "UASC": "0"},  # 6
        ]
    )
    fake_header_last = pd.DataFrame(
        [
            {"CHILD": "111", "UASC": "0"},  # 0
            {"CHILD": "222", "UASC": "1"},  # 2
            {"CHILD": "333", "UASC": "0"},  # 4
            {"CHILD": "345", "UASC": "0"},  # 5
            {"CHILD": "444", "UASC": "1"},  # 6
        ]
    )

    fake_dfs = {
        "Episodes": fake_episodes,
        "Header": fake_header,
        "Header_last": fake_header_last,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 2, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1008",
    message="Ofsted Unique Reference Number (URN) is not valid.",
    affected_fields=["URN"],
    tables=["Episodes", "Provider Info"],
)
def validate(dfs):
    if "Episodes" not in dfs or "provider_info" not in dfs["metadata"]:
        return {}
    else:
        episodes = dfs["Episodes"]
        providers = dfs["metadata"]["provider_info"]

        episodes["index_eps"] = episodes.index
        episodes = episodes[
            episodes["URN"].notna() & (episodes["URN"] != "XXXXXXX")
        ].copy()
        episodes["URN"] = episodes["URN"].astype(str)
        episodes = episodes.merge(providers, on="URN", how="left", indicator=True)
        mask = episodes["_merge"] == "left_only"
        eps_error_locations = episodes.loc[mask, "index_eps"]
        return {"Episodes": eps_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": "1111",
                "URN": "SC999999",
            },  # 0 pass
            {
                "CHILD": "1111",
                "URN": pd.NA,
            },  # 1 ignore
            {
                "CHILD": "1111",
                "URN": 1234567,
            },  # 2 pass: digits will be converted to strings before comparison.
            {
                "CHILD": "2222",
                "URN": "XXXXXXX",
            },  # 3 pass: accepted placeholder value
            {
                "CHILD": "3333",
                "URN": "1234567",
            },  # 4 pass
            {
                "CHILD": "3333",
                "URN": "2345",
            },  # 5 fail
            {
                "CHILD": "4444",
                "URN": "999999",
            },  # 6 pass
            {
                "CHILD": "5555",
                "URN": "5b67891",
            },  # 7 fail
            {
                "CHILD": "5555",
                "URN": "XXXXXX",
            },  # 8 fail: 6 Xs instead of seven
        ]
    )

    metadata = {
        "provider_info": pd.DataFrame({"URN": ["1234567", "SC999999", "999999"]})
    }
    fake_dfs = {"Episodes": fake_data_eps, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [5, 7, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="115",
    message="Date of Local Authority's (LA) decision that a child should be placed for adoption is not a valid date.",
    affected_fields=["DATE_PLACED"],
    tables=["PlacedAdoption"],
)
def validate(dfs):
    if "PlacedAdoption" not in dfs:
        return {}
    else:
        adopt = dfs["PlacedAdoption"]
        mask = pd.to_datetime(
            adopt["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        ).notna()

        na_location = adopt["DATE_PLACED"].isna()

        validation_error_mask = ~mask & ~na_location
        validation_error_locations = adopt.index[validation_error_mask]

        return {"PlacedAdoption": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DATE_PLACED": [
                "01/01/2021",
                "19/02/2010",
                "38/04/2019",
                "01/01/19",
                pd.NA,
            ],
        }
    )

    fake_dfs = {"PlacedAdoption": fake_data}

    result = validate(fake_dfs)

    assert result == {"PlacedAdoption": [2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="601",
    message="The additional fields relating to adoption have not been completed although the episode data shows that the child was adopted during the year.",
    affected_fields=[
        "REC",
        "DATE_INT",
        "DATE_MATCH",
        "FOSTER_CARE",
        "NB_ADOPTR",
        "SEX_ADOPTR",
        "LS_ADOPTR",
    ],
    tables=["Episodes", "AD1"],
)
def validate(dfs):
    if "Episodes" not in dfs or "AD1" not in dfs:
        return {}
    else:
        ad1 = dfs["AD1"]
        episodes = dfs["Episodes"]
        collection_start = dfs["metadata"]["collection_start"]
        collection_end = dfs["metadata"]["collection_end"]

        # prepare to merge
        ad1.reset_index(inplace=True)
        episodes.reset_index(inplace=True)

        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )

        # only keep episodes with adoption RECs during year
        adoption_eps_mask = (
            (episodes["DEC"] >= collection_start)
            & (episodes["DEC"] <= collection_end)
            & episodes["REC"].isin(["E11", "E12"])
        )
        episodes = episodes[adoption_eps_mask]

        # inner merge to take only episodes of children which are also found in the ad1 table
        merged = episodes.merge(ad1, on="CHILD", how="inner", suffixes=["_eps", "_ad1"])

        some_absent = (
            merged[
                [
                    "DATE_INT",
                    "DATE_MATCH",
                    "FOSTER_CARE",
                    "NB_ADOPTR",
                    "SEX_ADOPTR",
                    "LS_ADOPTR",
                ]
            ]
            .isna()
            .any(axis=1)
        )

        error_locs_ad1 = merged.loc[some_absent, "index_ad1"].unique().tolist()
        error_locs_eps = merged.loc[some_absent, "index_eps"].unique().tolist()

        return {"AD1": error_locs_ad1, "Episodes": error_locs_eps}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        [
            {
                "CHILD": "101",
                "DEC": "01/01/2009",
                "REC": "E45",
                "DECOM": "01/01/2009",
            },  # 0
            {
                "CHILD": "102",
                "DEC": "01/01/2021",
                "REC": "E11",
                "DECOM": "01/01/2001",
            },  # 1 fail
            {
                "CHILD": "102",
                "DEC": "20/12/2021",
                "REC": "E15",
                "DECOM": "20/12/2001",
            },  # 2
            {
                "CHILD": "102",
                "DEC": "03/07/2020",
                "REC": "E12",
                "DECOM": "03/01/2019",
            },  # 3 fail
            {
                "CHILD": "102",
                "DEC": "03/04/2019",
                "REC": "E48",
                "DECOM": "03/04/2008",
            },  # 4
            {
                "CHILD": "103",
                "DEC": "26/05/2020",
                "REC": "E12",
                "DECOM": "01/01/2002",
            },  # 5 pass
            {
                "CHILD": "104",
                "DEC": "10/01/2002",
                "REC": "E11",
                "DECOM": "10/01/2002",
            },  # 6
            {
                "CHILD": "104",
                "DEC": "11/02/2010",
                "REC": "X1",
                "DECOM": "11/02/2010",
            },  # 7
            {
                "CHILD": "104",
                "DEC": "25/01/2002",
                "REC": "X1",
                "DECOM": "25/01/2002",
            },  # 8
            {
                "CHILD": "105",
                "DEC": "25/01/2002",
                "REC": "E47",
                "DECOM": "25/01/2002",
            },  # 9
            {"CHILD": "105", "DEC": pd.NA, "REC": "E45", "DECOM": pd.NA},  # 10
        ]
    )
    fake_data_ad1 = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105"],
            "DATE_INT": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "DATE_MATCH": [pd.NA, "XXX", "XXX", pd.NA, "XXX"],
            "FOSTER_CARE": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "NB_ADOPTR": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "SEX_ADOPTR": [pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "LS_ADOPTR": [pd.NA, pd.NA, "XXX", "XXX", "XXX"],
        }
    )
    metadata = {"collection_start": "01/04/2020", "collection_end": "31/03/2021"}
    fake_dfs = {
        "Episodes": fake_data_episodes,
        "AD1": fake_data_ad1,
        "metadata": metadata,
    }

    result = validate(fake_dfs)

    assert result == {
        "AD1": [
            1,
        ],
        "Episodes": [1, 3],
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="364",
    message="Sections 41-46 of Police and Criminal Evidence (PACE; 1984) severely limits "
    + "the time a child can be detained in custody in Local Authority (LA) accommodation.",
    affected_fields=["LS", "DECOM", "DEC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    episodes = dfs["Episodes"]
    collection_end_str = dfs["metadata"]["collection_end"]

    J2_eps = episodes[episodes["LS"] == "J2"].copy()
    J2_eps["original_index"] = J2_eps.index

    J2_eps["DECOM"] = pd.to_datetime(
        J2_eps["DECOM"], format="%d/%m/%Y", errors="coerce"
    )
    J2_eps = J2_eps[J2_eps["DECOM"].notna()]
    J2_eps.loc[J2_eps["DEC"].isna(), "DEC"] = collection_end_str
    J2_eps["DEC"] = pd.to_datetime(J2_eps["DEC"], format="%d/%m/%Y", errors="coerce")

    J2_eps.sort_values(["CHILD", "DECOM"])

    J2_eps["index"] = pd.RangeIndex(0, len(J2_eps))
    J2_eps["index_prev"] = J2_eps["index"] + 1
    J2_eps = J2_eps.merge(
        J2_eps,
        left_on="index",
        right_on="index_prev",
        how="left",
        suffixes=[None, "_prev"],
    )
    J2_eps = J2_eps[
        ["original_index", "DECOM", "DEC", "DEC_prev", "CHILD", "CHILD_prev", "LS"]
    ]

    J2_eps["new_period"] = (J2_eps["DECOM"] > J2_eps["DEC_prev"]) | (
        J2_eps["CHILD"] != J2_eps["CHILD_prev"]
    )

    J2_eps["duration"] = (J2_eps["DEC"] - J2_eps["DECOM"]).dt.days
    J2_eps["period_id"] = J2_eps["new_period"].astype(int).cumsum()
    J2_eps["period_duration"] = J2_eps.groupby("period_id")["duration"].transform(sum)

    error_mask = J2_eps["period_duration"] > 21

    return {"Episodes": J2_eps.loc[error_mask, "original_index"].to_list()}


def test_validate():
    import pandas as pd

    test_eps = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DECOM": "01/01/2000",
                "DEC": "01/01/2001",
                "LS": "J2",
            },  # 0 Fail!
            {"CHILD": 101, "DECOM": "01/01/2001", "DEC": "20/12/2001", "LS": "X"},  # 1
            {
                "CHILD": 101,
                "DECOM": "20/12/2001",
                "DEC": "12/01/2002",
                "LS": "J2",
            },  # 2 Fail!
            {
                "CHILD": 2002,
                "DECOM": "01/01/2000",
                "DEC": "10/01/2000",
                "LS": "J2",
            },  # 3 ^ Fa
            {
                "CHILD": 2002,
                "DECOM": "10/01/2000",
                "DEC": "23/01/2001",
                "LS": "J2",
            },  # 4 v il!
            {
                "CHILD": 2002,
                "DECOM": "01/01/2002",
                "DEC": "10/01/2002",
                "LS": "J2",
            },  # 5
            {"CHILD": 2002, "DECOM": "10/01/2002", "DEC": "11/01/2002", "LS": "X"},  # 6
            {
                "CHILD": 2002,
                "DECOM": "11/01/2002",
                "DEC": "25/01/2002",
                "LS": "J2",
            },  # 7
            {
                "CHILD": 30003,
                "DECOM": "25/01/2002",
                "DEC": "10/03/2001",
                "LS": "J2",
            },  # 8 DEC < DECOM
            {
                "CHILD": 30003,
                "DECOM": "25/01/2002",
                "DEC": pd.NA,
                "LS": "J2",
            },  # 9 Fail!
            {
                "CHILD": 30003,
                "DECOM": pd.NA,
                "DEC": "25/01/2002",
                "LS": "J2",
            },  # 10 decom.isNA
            {"CHILD": 30003, "DECOM": pd.NA, "DEC": pd.NA, "LS": "J2"},  # 11
        ]
    )

    test_meta = {"collection_end": "01/04/2002"}

    test_dfs = {"Episodes": test_eps, "metadata": test_meta}

    result = validate(test_dfs)

    assert result == {"Episodes": [0, 2, 3, 4, 9]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW11aSTG2",
    message="Social worker episode has ended before 31 March and no subsequent social worker episode has been provided.",
    affected_fields=["SW_DEC"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]

        df["SW_DEC"] = pd.to_datetime(df["SW_DEC"], format="%d/%m/%Y", errors="coerce")
        df["SW_DECOM"] = pd.to_datetime(
            df["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        collection_end = dfs["metadata"]["collection_end"]
        collection_end = pd.to_datetime(collection_end, format="%d/%m/%Y")

        df["index"] = df.index
        df = df.sort_values(["CHILD", "SW_DECOM"])

        df_lead = df.shift(1)
        df_lead = df_lead.reset_index()

        m_df = df.merge(
            df_lead,
            left_on="index",
            right_on="level_0",
            how="outer",
            suffixes=("", "_prev"),
        )

        before_end = m_df[m_df["SW_DEC_prev"] <= collection_end - pd.DateOffset(days=5)]
        decom_within_timeframe = (
            before_end["SW_DECOM"] >= before_end["SW_DEC_prev"] - pd.DateOffset(days=5)
        ) & (
            before_end["SW_DECOM"] <= before_end["SW_DEC_prev"] + pd.DateOffset(days=5)
        )

        same_child = before_end["CHILD"] == before_end["CHILD_prev"]

        error_rows = before_end[
            (before_end["SW_DECOM"].isna() | ~decom_within_timeframe) & same_child
        ]["index"]

        return {"SWEpisodes": error_rows.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "1", "SW_DECOM": "01/04/2000", "SW_DEC": "25/03/2000"},
            {"CHILD": "1", "SW_DECOM": pd.NA, "SW_DEC": pd.NA},
            {"CHILD": "2", "SW_DECOM": "01/04/2000", "SW_DEC": "25/03/2000"},
            {"CHILD": "2", "SW_DECOM": "02/04/2000", "SW_DEC": "03/04/2000"},
            {"CHILD": "3", "SW_DECOM": "01/04/2000", "SW_DEC": "1/04/2000"},
            {"CHILD": "3", "SW_DECOM": "02/04/2000", "SW_DEC": "03/04/2000"},
            {"CHILD": "4", "SW_DECOM": "01/04/2000", "SW_DEC": "30/03/2000"},
            {"CHILD": "4", "SW_DECOM": "30/03/2000", "SW_DEC": "03/04/2000"},
            {"CHILD": "5", "SW_DECOM": "01/04/2000", "SW_DEC": "30/03/2000"},
            {
                "CHILD": "5",
                "SW_DECOM": pd.NA,
                "SW_DEC": pd.NA,
            },  # passes, SW_DEC later than end - 5
            {"CHILD": "6", "SW_DECOM": "01/04/2000", "SW_DEC": "25/03/2000"},
            {
                "CHILD": "6",
                "SW_DECOM": "29/03/2000",
                "SW_DEC": "03/04/2000",
            },  # passes, SW_DECOM within timeframe
            {"CHILD": "7", "SW_DECOM": "01/04/2000", "SW_DEC": "20/03/2000"},
            {"CHILD": "7", "SW_DECOM": "21/03/2000", "SW_DEC": "22/03/2000"},
            {
                "CHILD": "7",
                "SW_DECOM": "29/03/2000",
                "SW_DEC": "03/04/2000",
            },  # Fails, not within timeframe, child 7 checks it works for children with multiple rows
        ]
    )

    fake_meta = {"collection_end": "31/3/2000"}

    fake_dfs = {"SWEpisodes": fake_data, "metadata": fake_meta}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [1, 3, 14]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="544",
    message="Any child who has conviction information completed must also have immunisation, teeth check, health assessment and substance misuse problem identified fields completed.",
    affected_fields=[
        "CONVICTED",
        "IMMUNISATIONS",
        "TEETH_CHECK",
        "HEALTH_ASSESSMENT",
        "SUBSTANCE_MISUSE",
    ],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]

        convict = oc2["CONVICTED"].astype(str) == "1"
        immunisations = oc2["IMMUNISATIONS"].isna()
        teeth_ck = oc2["TEETH_CHECK"].isna()
        health_ass = oc2["HEALTH_ASSESSMENT"].isna()
        sub_misuse = oc2["SUBSTANCE_MISUSE"].isna()

        error_mask = convict & (immunisations | teeth_ck | health_ass | sub_misuse)
        validation_error_locations = oc2.index[error_mask]

        return {"OC2": validation_error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CONVICTED": [pd.NA, 1, "1", 1, "1", 1, "1", 1, "1", 1],
            "IMMUNISATIONS": [
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                1,
            ],
            "TEETH_CHECK": [
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "1",
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                1,
            ],
            "HEALTH_ASSESSMENT": [
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "1",
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                1,
            ],
            "SUBSTANCE_MISUSE": [
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                1,
            ],
        }
    )

    fake_dfs = {"OC2": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC2": [1, 2, 3, 4, 5, 6, 7, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="345",
    message="The data collection record shows the local authority is in touch with this young person, but activity and/or accommodation data items are zero.",
    affected_fields=["IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["OC3"],
)
def validate(dfs):
    if "OC3" not in dfs:
        return {}
    else:
        oc3 = dfs["OC3"]
        # If <IN_TOUCH> = 'Yes' then <ACTIV> and <ACCOM> must not be 0
        mask = (oc3["IN_TOUCH"] == "YES") & (
            (oc3["ACTIV"].astype(str) == "0") | (oc3["ACCOM"].astype(str) == "0")
        )
        error_locations = oc3.index[mask]
        return {"OC3": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data_oc3 = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C", "D", "E"],
            "IN_TOUCH": ["No", "YES", "YES", pd.NA, "Yes"],
            "ACTIV": [pd.NA, 0, "XXX", pd.NA, "XXX"],
            "ACCOM": [0, pd.NA, "0", "XXX", "XXX"],
        }
    )
    fake_dfs = {"OC3": fake_data_oc3}

    result = validate(fake_dfs)
    assert result == {"OC3": [1, 2]}



import pandas as pd
import numpy as np

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW10STG2",
    message="A new social worker episode has started before the end date of the previous social worker episode.",
    affected_fields=["SW_DECOM"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]

        df = df.reset_index()
        df["SW_DECOM"] = pd.to_datetime(
            df["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        df["SW_DEC"] = pd.to_datetime(df["SW_DEC"], format="%d/%m/%Y", errors="coerce")

        df = df.sort_values(["CHILD", "SW_DECOM", "SW_DEC"])

        df_lead = df.shift(1)
        df_lead = df_lead.reset_index()

        m_df = df.merge(
            df_lead,
            left_on=["CHILD", "index"],
            right_on=["CHILD", "level_0"],
            suffixes=("", "_prev"),
        )

        # <SW_DECOM> of current episode must be = (<SW_DEC>+/- 7 days) of previous episode if present.
        # Can be interpreted as meaning >= sw_dec - 7 days (which includes all days inside the tolerance)
        error_cohort = m_df[
            (~(m_df["SW_DECOM"] >= (m_df["SW_DEC_prev"] - np.timedelta64(7, "D"))))
            | (~(m_df["SW_DECOM"] <= (m_df["SW_DEC_prev"] + np.timedelta64(7, "D"))))
        ]
        error_list = error_cohort["index"].to_list()
        error_list.sort()

        return {"SWEpisodes": error_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "SW_DECOM": "01/01/2000",
                "SW_DEC": "09/01/2000",
            },  # 0 Ignore no previous SW episode
            {
                "CHILD": "1",
                "SW_DECOM": "02/01/2000",
                "SW_DEC": "03/01/2000",
            },  # 1 Pass >= previous SW_DEC (-7)
            {
                "CHILD": "2",
                "SW_DECOM": "01/01/1999",
                "SW_DEC": "09/01/2000",
            },  # 2 Ignore no previous SW episode
            {
                "CHILD": "2",
                "SW_DECOM": "01/01/2000",
                "SW_DEC": "10/01/2000",
            },  # 3 Fail not >= previous SW_DEC (-7)
            {
                "CHILD": "3",
                "SW_DECOM": "10/01/2000",
                "SW_DEC": "20/01/2000",
            },  # 4 Pass (technically) same as previous SW_DEC
            {
                "CHILD": "3",
                "SW_DECOM": "10/01/2000",
                "SW_DEC": "11/01/2000",
            },  # 5 Pass (technically) same as previous SW_DEC
            {
                "CHILD": "3",
                "SW_DECOM": "18/01/2000",
                "SW_DEC": "20/01/2000",
            },  # 6 Pass <= previous SW_DEC (+7) (accounting for SW_DEC in sort)
            {
                "CHILD": "3",
                "SW_DECOM": "30/01/2000",
                "SW_DEC": "31/01/2000",
            },  # 7 Fail >= previous SW_DEC (+7) (accounting for SW_DEC in sort)
        ]
    )

    fake_dfs = {"SWEpisodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [3, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="634",
    message="There are entries for previous permanence options, but child has not started to be looked after from 1 April 2016 onwards.",
    affected_fields=["LA_PERM", "PREV_PERM", "DATE_PERM", "DECOM"],
    tables=["Episodes", "PrevPerm"],
)
def validate(dfs):
    if "Episodes" not in dfs or "PrevPerm" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        prevperm = dfs["PrevPerm"]
        collection_start = dfs["metadata"]["collection_start"]
        # convert date field to appropriate format
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )
        # the maximum date has the highest possibility of satisfying the condition
        episodes["LAST_DECOM"] = episodes.groupby("CHILD")["DECOM"].transform("max")

        # prepare to merge
        episodes.reset_index(inplace=True)
        prevperm.reset_index(inplace=True)
        merged = prevperm.merge(
            episodes, on="CHILD", how="left", suffixes=["_prev", "_eps"]
        )
        # If <PREV_PERM> or <LA_PERM> or <DATE_PERM> provided, then at least 1 episode must have a <DECOM> later than 01/04/2016
        mask = (
            merged["PREV_PERM"].notna()
            | merged["DATE_PERM"].notna()
            | merged["LA_PERM"].notna()
        ) & (merged["LAST_DECOM"] < collection_start)
        eps_error_locs = merged.loc[mask, "index_eps"]
        prevperm_error_locs = merged.loc[mask, "index_prev"]

        # return {'PrevPerm':prevperm_error_locs}
        return {
            "Episodes": eps_error_locs.unique().tolist(),
            "PrevPerm": prevperm_error_locs.unique().tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_prevperm = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "6", "7", "8"],
            "PREV_PERM": ["Z1", pd.NA, pd.NA, "Z1", pd.NA, "P1"],
            "LA_PERM": [pd.NA, "352", pd.NA, "352", pd.NA, "352"],
            "DATE_PERM": [pd.NA, pd.NA, "01/05/2000", pd.NA, pd.NA, "05/05/2020"],
            # last 3 rows will disappear after merging on CHILD
        }
    )
    fake_data_episodes = pd.DataFrame(
        [
            {
                "CHILD": "101",
                "DECOM": "20/10/2011",
            },  # 0
            {
                "CHILD": "102",
                "DECOM": "19/11/2021",
            },  # 1
            {
                "CHILD": "102",
                "DECOM": "17/06/2001",
            },  # 2
            {
                "CHILD": "103",
                "DECOM": "04/04/2002",
            },  # 3
            {
                "CHILD": "103",
                "DECOM": "11/09/2015",
            },  # 4
            {
                "CHILD": "103",
                "DECOM": "01/03/2016",
            },  # 5
        ]
    )
    fake_dfs = {
        "PrevPerm": fake_data_prevperm,
        "Episodes": fake_data_episodes,
        "metadata": {"collection_start": "01/04/2016"},
    }

    result = validate(fake_dfs)

    assert result == {"PrevPerm": [0, 2], "Episodes": [0, 3, 4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="335",
    message="The current foster value (0) suggests that child is not adopted by current foster carer, but last placement is A2, A3, or A5. Or the current foster value (1) suggests that child is adopted by current foster carer, but last placement is A1, A4 or A6.",
    affected_fields=["PLACE", "FOSTER_CARE"],
    tables=["Episodes", "AD1"],
)
def validate(dfs):
    if "Episodes" not in dfs or "AD1" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        ad1 = dfs["AD1"]

        # prepare to merge
        episodes.reset_index(inplace=True)
        ad1.reset_index(inplace=True)
        merged = episodes.merge(ad1, on="CHILD", how="left", suffixes=["_eps", "_ad1"])

        # Where <PL> = 'A2', 'A3' or 'A5' and <DEC> = 'E1', 'E11', 'E12' <FOSTER_CARE> should not be '0'; Where <PL> = ‘A1’, ‘A4’ or ‘A6’ and <REC> = ‘E1’, ‘E11’, ‘E12’ <FOSTER_CARE> should not be ‘1’.
        mask = merged["REC"].isin(["E1", "E11", "E12"]) & (
            (
                merged["PLACE"].isin(["A2", "A3", "A5"])
                & (merged["FOSTER_CARE"].astype(str) == "0")
            )
            | (
                merged["PLACE"].isin(["A1", "A4", "A6"])
                & (merged["FOSTER_CARE"].astype(str) == "1")
            )
        )
        eps_error_locs = merged.loc[mask, "index_eps"]
        ad1_error_locs = merged.loc[mask, "index_ad1"]

        # use .unique since join is many to one
        return {
            "Episodes": eps_error_locs.tolist(),
            "AD1": ad1_error_locs.unique().tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        {
            "PLACE": ["A2", "--", "A1", "--", "A4", "--", "A5", "A3"],
            "REC": ["E1", "E1", "E1", "E12", "E11", "E1", "--", "E12"],
            "CHILD": ["10", "11", "12", "13", "14", "15", "16", "17"],
        }
    )
    fake_data_ad1 = pd.DataFrame(
        {
            "FOSTER_CARE": [0, 1, "0", "1", 2, "former foster carer", "", pd.NA],
            "CHILD": ["10", "11", "12", "13", "14", "15", "16", "17"],
        }
    )
    fake_dfs = {"Episodes": fake_data_eps, "AD1": fake_data_ad1}

    result = validate(fake_dfs)
    assert result == {"Episodes": [0], "AD1": [0]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="357",
    message="If this is the first episode ever for this child, reason for new episode must be S.  "
    "Check whether there is an episode immediately preceding this one, which has been left out.  "
    "If not the reason for new episode code must be amended to S.",
    affected_fields=["RNE"],
    tables=["Episodes"],
)

# !# Potential false negatives for first episodes before the current collection year?
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    eps = dfs["Episodes"]
    collection_start = pd.to_datetime(
        dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
    )
    eps["DECOM"] = pd.to_datetime(eps["DECOM"], format="%d/%m/%Y", errors="coerce")

    eps = eps.loc[eps["DECOM"].notnull()]

    first_eps = eps.loc[eps.groupby("CHILD")["DECOM"].idxmin()].loc[
        eps["DECOM"] >= collection_start
    ]

    errs = first_eps[first_eps["RNE"] != "S"].index.to_list()

    return {"Episodes": errs}


def test_validate():
    import pandas as pd

    episodes = pd.DataFrame(
        [
            {"CHILD": "000", "DECOM": "01/01/2001", "RNE": "!!"},
            {"CHILD": "111", "DECOM": "11/11/2011", "RNE": "ok"},
            {"CHILD": "111", "DECOM": "01/01/2000", "RNE": "S"},
            {"CHILD": "222", "DECOM": "11/11/2011", "RNE": "ok"},
            {"CHILD": "222", "DECOM": pd.NA, "RNE": "ok"},
            {"CHILD": "222", "DECOM": "01/01/2001", "RNE": "S"},
            {"CHILD": "333", "DECOM": "12/12/2012", "RNE": "ok"},
            {"CHILD": "333", "DECOM": "01/01/2001", "RNE": "!!"},
            {"CHILD": "333", "DECOM": "05/05/2005", "RNE": "ok"},
            {"CHILD": "444", "DECOM": "05/05/2005", "RNE": pd.NA},  # !!
            {
                "CHILD": "555",
                "DECOM": "01/01/1990",
                "RNE": "ok",
            },  # ok - before collection_start
        ]
    )
    metadata = {"collection_start": "31/03/1995"}
    test_dfs = {"Episodes": episodes, "metadata": metadata}

    test_result = validate(test_dfs)

    assert test_result == {"Episodes": [0, 7, 9]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="514",
    message="Data entry on the legal status of adopters shows a single adopter but data entry for the numbers of adopters shows it as a couple.",
    affected_fields=["LS_ADOPTR", "SEX_ADOPTR"],
    tables=["AD1"],
)
def validate(dfs):
    if "AD1" not in dfs:
        return {}
    else:
        AD1 = dfs["AD1"]
        code_list = ["M1", "F1"]
        # Check if LS Adopter is L0 and Sex Adopter is not M1 or F1.
        error_mask = (AD1["LS_ADOPTR"] == "L0") & (~AD1["SEX_ADOPTR"].isin(code_list))

        error_locations = AD1.index[error_mask]

        return {"AD1": error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "LS_ADOPTR": ["L0", "xx", "L0", pd.NA, "L0"],
            "SEX_ADOPTR": ["M1", "F1", "xx", pd.NA, "xxx"],
        }
    )

    fake_dfs = {"AD1": fake_data}

    result = validate(fake_dfs)

    assert result == {
        "AD1": [
            2,
            4,
        ]
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="117",
    message="Date of decision that a child should/should no longer be placed for adoption is beyond the current collection year or after the child ceased to be looked after.",
    affected_fields=["DATE_PLACED_CEASED", "DATE_PLACED", "DEC", "REC", "DECOM"],
    tables=["Episodes", "PlacedAdoption"],
)
def validate(dfs):
    if "Episodes" not in dfs or "PlacedAdoption" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        placed_adoption = dfs["PlacedAdoption"]
        collection_end = dfs["metadata"]["collection_end"]

        # datetime
        placed_adoption["DATE_PLACED_CEASED"] = pd.to_datetime(
            placed_adoption["DATE_PLACED_CEASED"],
            format="%d/%m/%Y",
            errors="coerce",
        )
        placed_adoption["DATE_PLACED"] = pd.to_datetime(
            placed_adoption["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        )
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        # Drop nans and continuing episodes
        episodes = episodes.dropna(subset=["DECOM"])
        episodes = episodes[episodes["REC"] != "X1"]

        episodes = episodes.loc[episodes.groupby("CHILD")["DECOM"].idxmax()]

        # prepare to merge
        placed_adoption.reset_index(inplace=True)
        episodes.reset_index(inplace=True)

        p4a_cols = ["DATE_PLACED", "DATE_PLACED_CEASED"]

        # latest episodes
        merged = episodes.merge(
            placed_adoption, on="CHILD", how="left", suffixes=["_eps", "_pa"]
        )
        mask = (
            (merged["DATE_PLACED"] > collection_end)
            | (merged["DATE_PLACED"] > merged["DEC"])
            | (merged["DATE_PLACED_CEASED"] > collection_end)
            | (merged["DATE_PLACED_CEASED"] > merged["DEC"])
        )
        # If provided <DATE_PLACED> and/or <DATE_PLACED_CEASED> must not be > <COLLECTION_END_DATE> or <DEC> of latest episode where <REC> not = 'X1'
        pa_error_locs = merged.loc[mask, "index_pa"]
        eps_error_locs = merged.loc[mask, "index_eps"]
        return {
            "Episodes": eps_error_locs.tolist(),
            "PlacedAdoption": pa_error_locs.unique().tolist(),
        }


def test_validate():
    import pandas as pd

    metadata = {"collection_end": "31/03/2018"}
    fake_placed_adoption = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DATE_PLACED_CEASED": "26/05/2000",
                "DATE_PLACED": "26/05/2000",
            },  # 0
            {
                "CHILD": 102,
                "DATE_PLACED_CEASED": "01/07/2018",
                "DATE_PLACED": "26/05/2000",
            },
            # 1 Fail DATE_PLACED_CEASED > collection_end
            {
                "CHILD": 103,
                "DATE_PLACED_CEASED": "26/05/2000",
                "DATE_PLACED": pd.NA,
            },  # 2
            {
                "CHILD": 104,
                "DATE_PLACED_CEASED": "26/05/2017",
                "DATE_PLACED": "01/02/2016",
            },
            # 3 Fail greater than DEC of latest episode
            {
                "CHILD": 105,
                "DATE_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "26/05/2019",
            },  # 4 Fail DATE_PLACED > collection_end
        ]
    )
    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DEC": "01/01/2009",
                "REC": "E45",
                "DECOM": "01/01/2009",
            },  # 0
            {
                "CHILD": 102,
                "DEC": "01/01/2001",
                "REC": "A3",
                "DECOM": "01/01/2001",
            },  # 1
            {
                "CHILD": 102,
                "DEC": "20/12/2001",
                "REC": "E15",
                "DECOM": "20/12/2001",
            },  # 2
            {
                "CHILD": 102,
                "DEC": "03/01/2019",
                "REC": "E46",
                "DECOM": "03/01/2019",
            },  # 3 Fail
            {
                "CHILD": 102,
                "DEC": "03/04/2008",
                "REC": "E48",
                "DECOM": "03/04/2008",
            },  # 4
            {
                "CHILD": 103,
                "DEC": "01/01/2002",
                "REC": "X2",
                "DECOM": "01/01/2002",
            },  # 5
            {
                "CHILD": 104,
                "DEC": "10/01/2002",
                "REC": "E11",
                "DECOM": "10/01/2002",
            },  # 6
            {
                "CHILD": 104,
                "DEC": "11/02/2010",
                "REC": "X1",
                "DECOM": "11/02/2010",
            },  # 7 Fail Ignored
            {
                "CHILD": 104,
                "DEC": "25/01/2002",
                "REC": "X1",
                "DECOM": "25/01/2002",
            },  # 8 Ignored REC is X1
            {
                "CHILD": 105,
                "DEC": "25/01/2002",
                "REC": "E47",
                "DECOM": "25/01/2002",
            },  # 9
            {"CHILD": 105, "DEC": pd.NA, "REC": "E45", "DECOM": pd.NA},  # 10
        ]
    )
    # TODO: in  the scenario where the REC of the latest episodes is X1, should the episode before the lastest
    #  be considered instead?. This will entail filtering by X1 before doing idxmax. Is this what this rule means?.

    fake_dfs = {
        "Episodes": fake_data_eps,
        "metadata": metadata,
        "PlacedAdoption": fake_placed_adoption,
    }

    result = validate(fake_dfs)
    assert result == {"Episodes": [3, 6, 9], "PlacedAdoption": [1, 3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="359",
    message="Child being looked after following 18th birthday must be accommodated under section 20(5) of the Children Act 1989 in a community home.",
    affected_fields=["DEC", "LS", "PLACE"],
    tables=["Header", "Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Header" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        hea = dfs["Header"]
        hea["DOB"] = pd.to_datetime(hea["DOB"], format="%d/%m/%Y", errors="coerce")
        collection_end = pd.to_datetime(
            dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
        )

        epi.reset_index(inplace=True)
        epi = epi.merge(hea, on="CHILD", how="left", suffixes=["", "_HEA"])

        mask_older_18 = (epi["DOB"] + pd.offsets.DateOffset(years=18)) < collection_end
        mask_null_dec = epi["DEC"].isna()
        mask_is_V2_K2 = (epi["LS"] == "V2") & (epi["PLACE"] == "K2")

        error_mask = mask_older_18 & mask_null_dec & ~mask_is_V2_K2
        error_list = epi["index"][error_mask].to_list()
        error_list = list(set(error_list))

        return {"Episodes": error_list}


def test_validate():
    import pandas as pd

    fake_hea = pd.DataFrame(
        [
            {"CHILD": "111", "DOB": "01/06/2020"},
            {"CHILD": "222", "DOB": "05/06/2000"},
            {"CHILD": "333", "DOB": "05/06/2000"},
            {"CHILD": "444", "DOB": "06/06/2000"},
            {"CHILD": "555", "DOB": "06/06/2019"},
        ]
    )

    fake_epi = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DEC": "01/06/2020",
                "LS": "R1",
                "PLACE": "R2",
            },  # 0 DOB 01/06/2020
            {"CHILD": "222", "DEC": pd.NA, "LS": "V2", "PLACE": "K2"},
            # 1 DOB 05/06/2000 Passes older than 18, but has V2 K2
            {
                "CHILD": "333",
                "DEC": pd.NA,
                "LS": "V2",
                "PLACE": "K1",
            },  # 2 DOB 05/06/2000 Fails
            {
                "CHILD": "444",
                "DEC": pd.NA,
                "LS": "R1",
                "PLACE": "R2",
            },  # 3 DOB 06/06/2000 Fails
            {
                "CHILD": "555",
                "DEC": pd.NA,
                "LS": "R1",
                "PLACE": "R2",
            },  # 4 DOB 06/06/2019 Passes, Too young
        ]
    )

    metadata = {"collection_end": "31/03/2021"}

    fake_dfs = {"Header": fake_hea, "Episodes": fake_epi, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 3]}



import datetime
import json
import logging
from typing import Optional

from prpc_python import RpcApp

from lac_validator import lac_validator
from lac_validator.ingress import read_from_text
from lac_validator.report import Report
from lac_validator.rules.ruleset_utils import get_year_ruleset
from lac_validator.utils import process_uploaded_files

logger = logging.getLogger(__name__)
handler = logging.FileHandler(
    datetime.datetime.now().strftime("lac validator --%d-%m-%Y %H.%M.%S.log")
)

f_format = logging.Formatter("%(asctime)s - %(levelname)s - % (message)s")
handler.setFormatter(f_format)
logger.addHandler(handler)

app = RpcApp("validate_lac")


@app.call
def get_rules(collection_year: str) -> str:
    """
    :param str ruleset: validation year e.g "2023" for 2022/2023 validation rules.
    :return rules_df: available rule codes and definitions according to chosen ruleset.
    """
    ruleset_registry = get_year_ruleset(collection_year)

    rules = []
    for _, rule in ruleset_registry.items():
        rules.append(
            {
                "code": str(rule.code),
                "description": str(rule.code) + " - " + str(rule.message),
            }
        )

    return json.dumps(rules)


@app.call
def generate_tables(lac_data: dict) -> dict[str, dict]:
    """
    :param lac_data: files uploaded by user mapped to the field where files were uploaded.
    :return lac_data_tables:  a dictionary of dataframes that has been converted to json.
    """
    files_list = process_uploaded_files(lac_data)

    data_files, _ = read_from_text(files_list)

    # what the frontend will display
    lac_data_tables = {
        table_name: table_df.to_json(orient="records")
        for table_name, table_df in data_files.items()
    }

    return lac_data_tables


@app.call
def lac_validate(
    lac_data: dict,
    file_metadata: dict,
    selected_rules: Optional[list[str]] = None,
):
    """
    :param lac_data: keys are table names and values are LAC csv files.
    :param file_metadata: contains collection year and local authority as strings.
    :param selected_rules: array of rules the user has chosen. consists of rule codes as strings.

    :return issue_report: issue locations in the data.
    :return rule_defs: codes and descriptions of the rules that triggers issues in the data.
    """
    # p4a_path = "tests\\fake_data\\placed_for_adoption_errors.csv"
    # ad1_path = "tests\\fake_data\\ad1.csv"
    # file_metadata = {"collectionYear": "2022", "localAuthority": "E09000027"}
    # lac_data = {"This year":[p4a_path, ad1_path], "Prev year": [p4a_path]}

    files_list = process_uploaded_files(lac_data)
    ruleset_registry = get_year_ruleset(file_metadata["collectionYear"])

    v = lac_validator.LacValidator(
        metadata=file_metadata,
        files=files_list,
        registry=ruleset_registry,
        selected_rules=selected_rules,
    )
    results = v.ds_results
    r = Report(results, ruleset_registry)
    full_issue_df = lac_validator.create_issue_df(r.report, r.error_report)

    # what the frontend will display
    issue_report = full_issue_df.to_json(orient="records")
    lac_data_tables = {
        table_name: table_df.to_json(orient="records")
        for table_name, table_df in v.dfs.items()
    }

    # what the user will download
    # r.error_summary generates the ErrorCounts file and r.child_summary generates the ChildErrorSummary file.
    user_reports = [
        r.error_summary.to_json(orient="records"),
        r.child_summary.to_json(orient="records"),
    ]

    validation_results = {
        "issue_locations": [issue_report],
        "data_tables": [lac_data_tables],
        "user_report": user_reports,
    }
    return validation_results



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.fixtures import current_episodes, previous_episodes

import pandas as pd


@rule_definition(
    code="453",
    message="Contradiction between placement distance in the last episode of the previous year and in the first episode of the current year.",
    affected_fields=["PL_DISTANCE"],
    tables=["Episodes", "Episodes_last"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Episodes_last" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi_last = dfs["Episodes_last"]

        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi_last["DECOM"] = pd.to_datetime(
            epi_last["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        epi_last["DEC"] = pd.to_datetime(
            epi_last["DEC"], format="%d/%m/%Y", errors="coerce"
        )

        epi.reset_index(inplace=True)

        first_ep_inds = epi.groupby(["CHILD"])["DECOM"].idxmin(skipna=True)
        min_decom = epi.loc[first_ep_inds, :]

        last_ep_inds = epi_last.groupby(["CHILD"])["DECOM"].idxmax(skipna=True)
        max_last_decom = epi_last.loc[last_ep_inds, :]

        merged_co = min_decom.merge(
            max_last_decom, how="inner", on=["CHILD"], suffixes=["", "_PRE"]
        )

        err_mask = (
            abs(
                merged_co["PL_DISTANCE"].astype(float, errors="ignore")
                - merged_co["PL_DISTANCE_PRE"].astype(float, errors="ignore")
            )
            >= 0.2
        ) | (
            merged_co["PL_DISTANCE_PRE"].isna() & merged_co["PL_DISTANCE"].notna()
            | (merged_co["PL_DISTANCE_PRE"].isna() & merged_co["PL_DISTANCE"].notna())
        )

        same_rne = merged_co["RNE"] == merged_co["RNE_PRE"]

        err_mask = err_mask & merged_co["DEC_PRE"].isna() & same_rne

        err_list = merged_co["index"][err_mask].unique().tolist()
        err_list.sort()
        return {"Episodes": err_list}


def test_validate():
    fake_dfs = {"Episodes": current_episodes, "Episodes_last": previous_episodes}
    result = validate(fake_dfs)

    assert result == {"Episodes": [4, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="399",
    message="Mother field, review field or participation field are completed but "
    + "child is looked after under legal status V3 or V4.",
    affected_fields=["MOTHER", "LS", "REVIEW", "REVIEW_CODE"],
    tables=["Episodes", "Header", "Reviews"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Header" not in dfs or "Reviews" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        header = dfs["Header"]
        reviews = dfs["Reviews"]

        code_list = ["V3", "V4"]
        # Rule adjustment: if child has any other episodes where LS is not V3 or V4, rule should not be triggered. Trigger error 399 only if all child episodes

        # Column that will contain True if LS of the episode is either V3 or V4
        episodes["LS_CHECKS"] = episodes["LS"].isin(code_list)

        # Column that will contain True only if all LSs, for a child, are either V3 or V4
        episodes["LS_CHECK"] = episodes.groupby("CHILD")["LS_CHECKS"].transform("min")

        eps = episodes.loc[episodes["LS_CHECK"] == True].copy()

        # prepare to merge
        eps["index_eps"] = eps.index
        header["index_hdr"] = header.index
        reviews["index_revs"] = reviews.index

        # merge
        merged = eps.merge(header, on="CHILD", how="left").merge(
            reviews, on="CHILD", how="left"
        )

        # If <LS> = 'V3' or 'V4' then <MOTHER>, <REVIEW> and <REVIEW_CODE> should not be provided
        mask = merged["LS_CHECK"] & (
            merged["MOTHER"].notna()
            | merged["REVIEW"].notna()
            | merged["REVIEW_CODE"].notna()
        )

        # Error locations
        eps_errors = merged.loc[mask, "index_eps"].dropna().unique()
        header_errors = merged.loc[mask, "index_hdr"].dropna().unique()
        revs_errors = merged.loc[mask, "index_revs"].dropna().unique()

        return {
            "Episodes": eps_errors.tolist(),
            "Header": header_errors.tolist(),
            "Reviews": revs_errors.tolist(),
        }


def test_validate():
    import pandas as pd

    # change log
    # index 6 and 7 are changed to the same child such that though both have the review information, index 6 would no longer trigger the error because the child has another episode (index 7) where LS is not V3/V4.
    # Index 0 and 1 have been same to the same child such that all its episodes have LS V3/V4 and the error is trigged though index zero does not have the review information and would not have triggered the error on its own.
    # index 6 truly reflects the change. It would have failed earlier but it passes now since the child's other LS is not V3
    # test data assumes that child IDs cannot repeat in the header and reviews tables.

    fake_data_episodes = pd.DataFrame(
        {
            "CHILD": ["101", "101", "103", "104", "105", "106", "108", "108"],
            "LS": ["V4", "V4", "V3", "L4", pd.NA, "V4", "V3", "XO"],
        }
    )
    fake_data_header = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "106", "108", "109"],
            "MOTHER": ["0", "0", pd.NA, pd.NA, 1, 0, 1, pd.NA],
        }
    )
    fake_data_reviews = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "106", "108", "109"],
            "REVIEW": [
                "31/12/2008",
                "01/01/2012",
                pd.NA,
                "01/01/2009",
                "01/01/2012",
                "01/01/2017",
                "01/01/2021",
                "01/01/2015",
            ],
            "REVIEW_CODE": ["PN1", "PN2", pd.NA, "PN4", "PN5", "PN6", "PN7", "PN0"],
        }
    )
    fake_dfs = {
        "Header": fake_data_header,
        "Reviews": fake_data_reviews,
        "Episodes": fake_data_episodes,
    }

    result = validate(fake_dfs)
    assert result == {
        "Header": [0, 5],
        "Reviews": [0, 5],
        "Episodes": [
            0,
            1,
            5,
        ],
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="516",
    message="The episode data submitted for this child does not show that he/she was with their former foster carer(s) during the year.If the code in the reason episode ceased is E45 or E46 the child must have a placement code of U1 to U6.",
    affected_fields=["REC", "PLACE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}

    else:
        episodes = dfs["Episodes"]
        place_codes = ["U1", "U2", "U3", "U4", "U5", "U6"]
        rec_codes = ["E45", "E46"]

        error_mask = episodes["REC"].isin(rec_codes) & ~episodes["PLACE"].isin(
            place_codes
        )

        validation_error_locations = episodes.index[error_mask]

        return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "REC": [
                "E45",
                "E46",
                pd.NA,
                "E45",
                "E46",
                "E45",
                "E45",
                "E2",
                "E2",
                "E46",
                "E46",
                "E46",
            ],
            "PLACE": [
                "U1",
                "U2",
                "U3",
                "Z1",
                "S1",
                "R1",
                "U4",
                "K2",
                "K2",
                "T1",
                "U6",
                "xx",
            ],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [3, 4, 5, 9, 11]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="556",
    message="Date of decision that the child should be placed for adoption should be on or prior to the date that the freeing order was granted.",
    affected_fields=["DATE_PLACED", "DECOM"],
    tables=["Episodes", "PlacedAdoption"],
)
def validate(dfs):
    if "Episodes" not in dfs or "PlacedAdoption" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        placedAdoptions = dfs["PlacedAdoption"]

        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        placedAdoptions["DATE_PLACED"] = pd.to_datetime(
            placedAdoptions["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        )

        episodes = episodes.reset_index()

        D1Episodes = episodes[episodes["LS"] == "D1"]

        merged = (
            D1Episodes.reset_index()
            .merge(
                placedAdoptions,
                how="left",
                on="CHILD",
            )
            .set_index("index")
        )

        episodes_with_errors = merged[merged["DATE_PLACED"] > merged["DECOM"]]

        error_mask = episodes.index.isin(episodes_with_errors.index)

        error_locations = episodes.index[error_mask]

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C"],
            "LS": ["C1", "D1", "D1"],
            "DECOM": [pd.NA, "15/03/2020", "15/03/2020"],
        }
    )
    fake_data_placed_adoption = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C"],
            "DATE_PLACED": [pd.NA, "15/04/2020", "15/02/2020"],
        }
    )

    fake_dfs = {
        "Episodes": fake_data_episodes,
        "PlacedAdoption": fake_data_placed_adoption,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="132",
    message="Data entry for activity after leaving care is invalid.",
    affected_fields=["ACTIV"],
    tables=["OC3"],
)
def validate(dfs):
    if "OC3" not in dfs:
        return {}

    care_leavers = dfs["OC3"]
    code_list = [
        "F1",
        "P1",
        "F2",
        "P2",
        "F4",
        "P4",
        "F5",
        "P5",
        "G4",
        "G5",
        "G6",
        "0",
    ]
    mask = care_leavers["ACTIV"].astype(str).isin(code_list)

    validation_error_mask = ~mask
    validation_error_locations = care_leavers.index[validation_error_mask]

    return {"OC3": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "ACTIV": ["f1", "F1", 1, 0, "1", "0", "", pd.NA],
        }
    )

    fake_dfs = {"OC3": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC3": [0, 2, 4, 6, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1001",
    message="The episodes recorded for this young person suggest they are not a relevant or a former relevant "
    "child and therefore should not have care leaver information completed. "
    "[NOTE: This tool can only test the current and previous year data loaded into the tool - this "
    "check may generate false positives if a child had episodes prior to last year's collection.]",
    affected_fields=["IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["Episodes", "OC3", "Header", "Episodes_last"],
)
def validate(dfs):
    # requiring 'Episodes_last' to reduce false positive rate, though more could be done
    if any(
        table_name not in dfs
        for table_name in ("Episodes", "OC3", "Header", "Episodes_last")
    ):
        return {}
    elif any(
        len(dfs[table_name]) == 0
        for table_name in ("Episodes", "OC3", "Header", "Episodes_last")
    ):
        return {}
    else:
        current_eps = dfs["Episodes"]
        prev_eps = dfs["Episodes_last"]
        oc3 = dfs["OC3"]
        header = dfs["Header"]

        collection_end = dfs["metadata"]["collection_end"]
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )

        episodes = pd.concat([current_eps, prev_eps], axis=0)
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        episodes.drop_duplicates(subset=["CHILD", "DECOM"])

        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        header = header[header["DOB"].notnull()]
        header["DOB14"] = header["DOB"] + pd.DateOffset(years=14)
        header["DOB16"] = header["DOB"] + pd.DateOffset(years=16)

        # Drop children who are over 20 years old at collection end,
        # as we would not expect to see sufficient episodes in the past 2 years of data
        header = header[header["DOB"] + pd.DateOffset(years=20) > collection_end]

        # this should drop any episodes duplicated between years.
        # keep='first' should drop prev. year's missing DEC
        episodes = episodes.sort_values("DEC").drop_duplicates(
            ["CHILD", "DECOM"], keep="first"
        )

        # fill in missing final DECs with the collection year's end date
        missing_last_DECs = (
            episodes.index.isin(episodes.groupby("CHILD")["DECOM"].idxmax())
            & episodes["DEC"].isna()
        )
        episodes.loc[missing_last_DECs, "DEC"] = collection_end

        # Work out how long child has been in care since 14th and 16th birthdays.
        episodes_merged = (
            episodes.reset_index()
            .merge(
                header[["CHILD", "DOB", "DOB14", "DOB16"]],
                how="inner",
                on=["CHILD"],
                suffixes=("", "_header"),
                indicator=True,
            )
            .set_index("index")
        )

        # Drop all episodes with V3/V4 legal status
        v3v4_ls = episodes_merged["LS"].str.upper().isin(["V3", "V4"])
        index_v3v4_ls = episodes_merged.loc[v3v4_ls].index
        episodes_merged.drop(index_v3v4_ls, inplace=True)

        if len(episodes_merged) == 0:
            return {"OC3": []}

        episodes_merged["DECOM14"] = episodes_merged[["DECOM", "DOB14"]].max(axis=1)
        episodes_merged["DECOM16"] = episodes_merged[["DECOM", "DOB16"]].max(axis=1)

        episodes_merged["DURATION14"] = (
            episodes_merged["DEC"] - episodes_merged["DECOM14"]
        ).dt.days.clip(lower=0)
        episodes_merged["DURATION16"] = (
            episodes_merged["DEC"] - episodes_merged["DECOM16"]
        ).dt.days.clip(lower=0)

        episodes_merged["TOTAL14"] = episodes_merged.groupby("CHILD")[
            "DURATION14"
        ].transform("sum")
        episodes_merged["TOTAL16"] = episodes_merged.groupby("CHILD")[
            "DURATION16"
        ].transform("sum")

        totals = episodes_merged[["CHILD", "TOTAL14", "TOTAL16"]].drop_duplicates(
            "CHILD"
        )

        oc3 = oc3.merge(totals, how="left")

        has_care_after_14 = oc3["TOTAL14"] >= 91
        has_care_after_16 = oc3["TOTAL16"] >= 1

        valid_care_leaver = has_care_after_14 & has_care_after_16

        # Find out if child has been adopted
        episodes_max = episodes.groupby("CHILD")["DECOM"].idxmax()
        episodes_max = episodes.loc[episodes_max]
        episodes_adopted = episodes_max[
            episodes_max["REC"].str.upper().isin(["E11", "E12"])
        ]
        adopted = oc3["CHILD"].isin(episodes_adopted["CHILD"])

        error_mask = adopted | ~valid_care_leaver

        validation_error_locations = oc3.index[error_mask]

        return {"OC3": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    # DOB always 01/01/2000
    # next to each episode, approx days in each age bracket is listed like so:      under 14  :  14-16  :  over 16
    eps = pd.DataFrame(
        [
            # [0] - PASS: more than 91 relevant days
            {
                "CHILD": "1111",
                "LS": "C2",
                "DECOM": "01/01/2014",
                "DEC": "01/02/2014",
                "REC": "o",
            },  # :30:
            {
                "CHILD": "1111",
                "LS": "C2",
                "DECOM": "01/02/2015",
                "DEC": "01/06/2015",
                "REC": "o",
            },  # :120:
            {
                "CHILD": "1111",
                "LS": "C2",
                "DECOM": "01/01/2016",
                "DEC": "01/02/2016",
                "REC": "o",
            },  # ::1
            # [1] - FAIL: less than 91 days
            {
                "CHILD": "2222",
                "LS": "C2",
                "DECOM": "01/01/2010",
                "DEC": "01/02/2014",
                "REC": "o",
            },  # :30:
            {
                "CHILD": "2222",
                "LS": "C2",
                "DECOM": "01/01/2010",
                "DEC": pd.NA,
                "REC": "o",
            },  # Duplicate DECOM, missing
            # DEC - should get dropped
            {
                "CHILD": "2222",
                "LS": "C2",
                "DECOM": "25/12/2015",
                "DEC": "01/02/2016",
                "REC": "o",
            },  # :7:30
            # [2] - FAIL: more than 91 days but not after 14th bday
            {
                "CHILD": "3333",
                "LS": "C2",
                "DECOM": "01/01/2010",
                "DEC": "01/03/2014",
                "REC": "o",
            },  # 120:60:
            {
                "CHILD": "3333",
                "LS": "C2",
                "DECOM": "25/12/2015",
                "DEC": "04/01/2016",
                "REC": "o",
            },  # :7:3
            # [3] - PASS: more than 91 days, all after 16th bday
            {
                "CHILD": "4444",
                "LS": "C2",
                "DECOM": "01/01/2016",
                "DEC": "01/02/2016",
                "REC": "o",
            },  # ::30
            {
                "CHILD": "4444",
                "LS": "C2",
                "DECOM": "25/12/2016",
                "DEC": "01/04/2017",
                "REC": "o",
            },  # ::120
            # [-] - PASS: not in OC3
            {
                "CHILD": "5555",
                "LS": "C2",
                "DECOM": "01/01/2010",
                "DEC": "01/05/2012",
                "REC": "o",
            },  # 120:60:
            {
                "CHILD": "5555",
                "LS": "C2",
                "DECOM": "25/12/2015",
                "DEC": "04/01/2016",
                "REC": "o",
            },  # :7:3
            # [4] - FAIL: more than 91 days but none after 16th bday
            {
                "CHILD": "6006",
                "LS": "C2",
                "DECOM": "01/01/2014",
                "DEC": "01/02/2014",
                "REC": "o",
            },  # :30:
            {
                "CHILD": "6006",
                "LS": "C2",
                "DECOM": "01/02/2015",
                "DEC": "01/08/2015",
                "REC": "o",
            },  # :180:
            # [5] - FAIL: enough days but including V3 episode
            {
                "CHILD": "7777",
                "LS": "C2",
                "DECOM": "01/01/2013",
                "DEC": "01/03/2014",
                "REC": "o",
            },  # 365:60:
            {
                "CHILD": "7777",
                "LS": "V3",
                "DECOM": "01/01/2015",
                "DEC": "01/02/2015",
                "REC": "o",
            },  # :30: (doesnt count)
            {
                "CHILD": "7777",
                "LS": "C2",
                "DECOM": "25/12/2015",
                "DEC": "04/01/2016",
                "REC": "o",
            },  # :7:3
            # [6] - FAIL:  more than 91 days but final REC E11/E12
            {
                "CHILD": "8888",
                "LS": "C2",
                "DECOM": "01/01/2014",
                "DEC": "01/02/2014",
                "REC": "o",
            },  # :30:
            {
                "CHILD": "8888",
                "LS": "C2",
                "DECOM": "01/02/2015",
                "DEC": "01/08/2015",
                "REC": "o",
            },  # :180:
            {
                "CHILD": "8888",
                "LS": "C2",
                "DECOM": "01/02/2016",
                "DEC": "01/03/2016",
                "REC": "o",
            },  # ::30
            {
                "CHILD": "8888",
                "LS": "C2",
                "DECOM": "01/03/2017",
                "DEC": "01/05/2017",
                "REC": "E11",
            },  # ::60 but E11
            # [8] - PASS: more than 91 days if missing DEC filled in
            {
                "CHILD": "9999",
                "LS": "C2",
                "DECOM": "01/01/2010",
                "DEC": "01/05/2012",
                "REC": "o",
            },  # 120:60:
            {
                "CHILD": "9999",
                "LS": "C2",
                "DECOM": "01/06/2017",
                "DEC": pd.NA,
                "REC": "o",
            },  # :40:120
        ]
    )

    metadata = {"collection_end": "01/10/2017"}

    # Split episodes into current and previous year dataframes
    eps["DECOM_dt"] = pd.to_datetime(eps["DECOM"], format="%d/%m/%Y", errors="coerce")
    first_ep_per_child = eps.groupby("CHILD")["DECOM_dt"].idxmin()
    prev_eps = eps.loc[first_ep_per_child].copy().drop("DECOM_dt", axis=1)
    current_eps = eps.drop(
        index=first_ep_per_child[:-4], columns="DECOM_dt"
    )  # keep a couple eps in both df's, so we know dups dont break it

    assert (
        len(prev_eps) + len(current_eps) == len(eps) + 4
    ), "(test logic problem) creating current/prev episodes tables didnt work as intended"

    oc3 = pd.DataFrame(
        {
            "CHILD": [
                "1111",
                "2222",
                "3333",
                "4444",
                "6006",
                "7777",
                "8888",
                "9999" "1010101010",
            ],  # '1010101010' not in episodes
        }
    )

    header = pd.DataFrame(
        {
            "CHILD": [
                "1111",
                "2222",
                "3333",
                "4444",
                "6006",
                "7777",
                "8888",
                "1010101010",
            ]
        }
    )
    header["DOB"] = "01/01/2000"

    fake_dfs = {
        "Episodes": current_eps,
        "OC3": oc3,
        "Header": header,
        "Episodes_last": prev_eps,
        "metadata": metadata,
    }
    result = validate(fake_dfs)
    assert result == {"OC3": [1, 2, 4, 5, 6, 7]}



import pandas as pd

from lac_validator.datastore import merge_postcodes
from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="392c",
    message="Postcode(s) provided are invalid.",
    affected_fields=["HOME_POST", "PL_POST"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]

        home_provided = episodes["HOME_POST"].notna()
        home_details = merge_postcodes(episodes, "HOME_POST")
        home_valid = home_details["pcd"].notna()

        pl_provided = episodes["PL_POST"].notna()
        pl_details = merge_postcodes(episodes, "PL_POST")
        pl_valid = pl_details["pcd"].notna()

        error_mask = (home_provided & ~home_valid) | (pl_provided & ~pl_valid)

        return {"Episodes": episodes.index[error_mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "HOME_POST": [
                "AB1 0JD",
                "invalid",
                "AB1 0JD",
                "invalid",
                "AB10JD",
            ],
            "PL_POST": [
                "AB1 0JD",
                "AB1 0JD",
                "invalid",
                "invalid",
                "AB1 0JD",
            ],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.types import MissingMetadataError


@rule_definition(
    code="205B",
    message="Child previously identified as UASC is also UASC this year, but date UASC ceased in both years does not support this.",
    affected_fields=["DUC", "UASC"],
    tables=["UASC", "UASC_last", "Header", "Header_last"],
)
def validate(dfs):
    try:
        file_format = dfs["metadata"]["file_format"]
    except KeyError as e:
        raise MissingMetadataError(*e.args)

    if "UASC" not in dfs or "UASC_last" not in dfs:
        return {}
    elif file_format == "xml" and ("Header" not in dfs or "Header_last" not in dfs):
        return {}
    elif all(i in dfs for i in ("Header", "Header_last", "UASC", "UASC_last")):
        return_header_errors = True

        uasc = dfs["UASC"]
        uasc_last = dfs["UASC_last"]
        header = dfs["Header"]
        header_last = dfs["Header_last"]
    elif file_format == "csv":
        # for csv uploads, the Header table gets the UASC column added in ingress if UASC is present,
        # as does Header_last if UASC_last is present.  Therefore use Header['UASC'] if present, else make our own
        uasc = dfs["UASC"]
        uasc_last = dfs["UASC_last"]
        if "Header" in dfs:
            return_header_errors = True

            header = dfs["Header"]
        else:
            return_header_errors = False

            header = uasc[["CHILD"]].copy()
            header["UASC"] = "0"
            uasc_inds = uasc.drop(["CHILD", "DOB"], axis="columns").notna().any(axis=1)
            header.loc[uasc_inds, "UASC"] = "1"
        if "Header_last" in dfs:
            header_last = dfs["Header_last"]
        else:
            header_last = uasc_last[["CHILD"]].copy()
            header_last["UASC"] = "0"
            uasc_inds = (
                uasc_last.drop(["CHILD", "DOB"], axis="columns").notna().any(axis=1)
            )
            header_last.loc[uasc_inds, "UASC"] = "1"
    else:
        raise RuntimeError("Table selection failed (205B). This shouldn't be possible.")
    if "UASC" not in header.columns or "UASC" not in header_last.columns:
        return {}
    collection_start = pd.to_datetime(
        dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
    )
    collection_end = pd.to_datetime(
        dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
    )
    collection_start_last = collection_start + pd.offsets.DateOffset(years=-1)
    collection_end_last = collection_end + pd.offsets.DateOffset(years=-1)

    uasc["DOB"] = pd.to_datetime(uasc["DOB"], format="%d/%m/%Y", errors="coerce")
    uasc["DUC"] = pd.to_datetime(uasc["DUC"], format="%d/%m/%Y", errors="coerce")
    uasc_last["DOB"] = pd.to_datetime(
        uasc_last["DOB"], format="%d/%m/%Y", errors="coerce"
    )
    uasc_last["DUC"] = pd.to_datetime(
        uasc_last["DUC"], format="%d/%m/%Y", errors="coerce"
    )
    uasc_last["DOB18"] = uasc_last["DOB"] + pd.offsets.DateOffset(years=18)

    header["HDR_INDEX"] = header.index
    header_last.reset_index(inplace=True)

    header.reset_index(inplace=True)
    uasc.reset_index(inplace=True)

    merged_current = uasc[["CHILD", "DUC", "index"]].merge(
        header[["CHILD", "UASC", "index"]],
        how="left",
        on="CHILD",
        suffixes=("_uasc", "_header"),
    )

    merged_last = uasc_last[["CHILD", "DUC", "DOB18"]].merge(
        header_last[["CHILD", "UASC"]], how="left", on="CHILD"
    )

    all_merged = merged_current.merge(
        merged_last,
        how="left",
        on=["CHILD"],
        suffixes=("", "_last"),
        indicator=True,
    )

    uasc_in_both_years = (all_merged["UASC"].astype(str) == "1") & (
        all_merged["UASC_last"].astype(str) == "1"
    )
    uasc_current_duc = (all_merged["DUC"] >= collection_start) & (
        all_merged["DUC"] <= collection_end
    )
    prev_duc_is_18th = all_merged["DUC_last"] == all_merged["DOB18"]
    current_duc_is_18th = all_merged["DUC"] == all_merged["DUC_last"]

    error_mask = uasc_in_both_years & (
        (~uasc_current_duc & ~current_duc_is_18th) | ~prev_duc_is_18th
    )

    error_locations_uasc = all_merged.loc[error_mask, "index_uasc"]
    error_locations_header = all_merged.loc[error_mask, "index_header"]

    if return_header_errors:
        return {
            "UASC": error_locations_uasc.to_list(),
            "Header": error_locations_header.to_list(),
        }
    else:
        return {"UASC": error_locations_uasc.to_list()}


def test_validate():
    import pandas as pd

    fake_uasc_205 = pd.DataFrame(
        [
            {"CHILD": "101", "DOB": "28/10/2004", "DUC": pd.NA},  # Pass C
            {"CHILD": "102", "DOB": "04/06/2004", "DUC": pd.NA},  # Pass A
            {"CHILD": "103", "DOB": "03/03/2002", "DUC": "10/07/2020"},  # Fail A
            {"CHILD": "104", "DOB": "28/03/2003", "DUC": "14/05/2021"},  # Fail A
            {"CHILD": "105", "DOB": "16/04/2001", "DUC": "16/04/2019"},  # Fail B
            {"CHILD": "106", "DOB": "04/11/2004", "DUC": "16/06/2021"},  # Fail B
            {"CHILD": "107", "DOB": "23/07/2002", "DUC": "23/07/2020"},  # Pass B
            {"CHILD": "108", "DOB": "19/02/2003", "DUC": pd.NA},  # Fail C
            {"CHILD": "109", "DOB": "14/06/2003", "DUC": "14/06/2021"},  # Fail D
            {"CHILD": "110", "DOB": "14/06/2003", "DUC": pd.NA},  # Fail D
        ]
    )

    fake_uasc_prev_205 = pd.DataFrame(
        [
            {"CHILD": "101", "DOB": "28/10/2004", "DUC": pd.NA},  # Pass C
            {"CHILD": "102", "DOB": "04/06/2004", "DUC": "20/01/2020"},  # Pass A
            {"CHILD": "103", "DOB": "03/03/2002", "DUC": "10/07/2020"},  # Fail A
            {"CHILD": "104", "DOB": "28/03/2003", "DUC": "14/05/2021"},  # Fail A
            {"CHILD": "105", "DOB": "16/04/2001", "DUC": "16/04/2019"},  # Fail B
            {"CHILD": "106", "DOB": "04/11/2004", "DUC": "04/11/2023"},  # Fail B
            {"CHILD": "107", "DOB": "23/07/2002", "DUC": "23/07/2020"},  # Pass B
            {"CHILD": "108", "DOB": "19/02/2003", "DUC": "19/02/2021"},  # Fail C
            {"CHILD": "109", "DOB": "14/06/2003", "DUC": "14/06/2021"},  # Fail D
            {"CHILD": "110", "DOB": "14/06/2003", "DUC": "14/06/2021"},  # Fail D
        ]
    )

    fake_header_205 = pd.DataFrame(
        [
            {"CHILD": "101", "DOB": "28/10/2004", "UASC": "0"},  # Pass C
            {"CHILD": "108", "DOB": "19/02/2003", "UASC": "0"},  # Fail C
            {"CHILD": "109", "DOB": "14/06/2003", "UASC": "1"},  # Fail D
            {"CHILD": "102", "DOB": "04/06/2004", "UASC": "0"},  # Pass A
            {"CHILD": "103", "DOB": "03/03/2002", "UASC": "0"},  # Fail A
            {"CHILD": "104", "DOB": "28/03/2003", "UASC": "0"},  # Fail A
            {"CHILD": "105", "DOB": "16/04/2001", "UASC": "1"},  # Fail B
            {"CHILD": "106", "DOB": "04/11/2004", "UASC": "1"},  # Fail B
            {"CHILD": "107", "DOB": "23/07/2002", "UASC": "1"},  # Pass B
            {"CHILD": "110", "DOB": "03/03/2002", "UASC": "0"},
        ]
    )

    prev_fake_header_205 = pd.DataFrame(
        [
            {"CHILD": "102", "DOB": "04/06/2004", "UASC": "1"},  # Pass A
            {"CHILD": "103", "DOB": "03/03/2002", "UASC": "1"},  # Fail A
            {"CHILD": "104", "DOB": "28/03/2003", "UASC": "1"},  # Fail A
            {"CHILD": "101", "DOB": "28/10/2004", "UASC": "0"},  # Pass C
            {"CHILD": "105", "DOB": "16/04/2001", "UASC": "1"},  # Fail B
            {"CHILD": "106", "DOB": "04/11/2004", "UASC": "1"},  # Fail B
            {"CHILD": "107", "DOB": "23/07/2002", "UASC": "1"},  # Pass B
            {"CHILD": "108", "DOB": "19/02/2003", "UASC": "0"},  # Fail C
            {"CHILD": "109", "DOB": "14/06/2003", "UASC": "0"},  # Fail D
            {"CHILD": "110", "DOB": "03/03/2002", "UASC": "0"},
        ]
    )

    metadata_205 = {
        "collection_start": "01/04/2020",
        "collection_end": "31/03/2021",
    }

    fake_dfs_205_xml = {
        "UASC": fake_uasc_205,
        "UASC_last": fake_uasc_prev_205,
        "Header": fake_header_205,
        "Header_last": prev_fake_header_205,
        "metadata": {**metadata_205, **{"file_format": "xml"}},
    }

    fake_dfs_205_csv_1 = {
        "UASC": fake_uasc_205,
        "UASC_last": fake_uasc_prev_205,
        "Header": fake_header_205,
        "metadata": {**metadata_205, **{"file_format": "csv"}},
    }

    fake_dfs_205_csv_2 = {
        "UASC": fake_uasc_205,
        "UASC_last": fake_uasc_prev_205,
        "metadata": {**metadata_205, **{"file_format": "csv"}},
    }

    dfs = {k: v.copy() for k, v in fake_dfs_205_xml.items()}
    result = validate(dfs)
    assert result == {"UASC": [5], "Header": [7]}

    dfs = {k: v.copy() for k, v in fake_dfs_205_csv_1.items()}
    result = validate(dfs)
    assert result == {"UASC": [5], "Header": [7]}

    dfs = {k: v.copy() for k, v in fake_dfs_205_csv_2.items()}
    result = validate(dfs)
    assert result == {"UASC": [2, 3, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="230",
    message="Ofsted Unique reference number (URN) is required for any English H5 placements from 28 October 2023.",
    affected_fields=["PLACE", "DEC", "URN"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        max_dec_allowed = pd.to_datetime(
            "29/10/2023", format="%d/%m/%Y", errors="coerce"
        )

        out_of_england = (
            episodes["PL_LA"].astype("str").str.upper().str.startswith(("N", "W", "S"))
        )

        mask = (
            episodes["PLACE"].isin(["H5"])
            & ((episodes["DEC"] >= max_dec_allowed) | episodes["DEC"].isna())
            & (episodes["URN"].isna() | (episodes["URN"] == "XXXXXXX"))
            & ~out_of_england
        )

        validation_error_mask = mask
        validation_error_locations = episodes.index[validation_error_mask]

        return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": ["H5", "H5", "H5", "H5", "P1", "H5", "H5"],
            "DEC": [
                "29/10/2023",
                "29/10/2023",
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "28/10/2023",
            ],
            "PL_LA": ["A1", "A1", "A1", "S1", "A1", "A1", "A1"],
            "URN": [pd.NA, "X1", pd.NA, pd.NA, pd.NA, "XXXXXXX", pd.NA],
            # 0 Fail (DEC on/after validation date and no URN)
            # 1 Pass (DEC on/after validation date with URN)
            # 2 Fail (Nil DEC and no URN)
            # 3 Ignore (Nil DEC and no URN *but* outside England)
            # 4 Ignore (Nil DEC and no URN *but* not H5)
            # 5 Fail (Nil DEC and URN is XXXXXXX)
            # 6 Pass (DEC before validation date and no URN)
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 2, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="301",
    message="Date of birth falls after the year ended.",
    affected_fields=["DOB"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        collection_end = dfs["metadata"]["collection_end"]

        # convert dates
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )
        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )

        # <DOB> must be <= <COLLECTION_END_DATE>
        mask = header["DOB"] > collection_end

        # error locations
        error_locs = header.index[mask]
        return {"Header": error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_data_header = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DOB": "01/07/2021",
            },  # 0 fail
            {
                "CHILD": 102,
                "DOB": "02/06/2000",
            },  # 1
            {
                "CHILD": 103,
                "DOB": "03/06/2000",
            },  # 2
            {
                "CHILD": 104,
                "DOB": "04/06/2022",
            },  # 3 fail
            {
                "CHILD": 105,
                "DOB": pd.NA,
            },  # 4
        ]
    )

    metadata = {"collection_start": "01/04/2020", "collection_end": "31/03/2021"}

    fake_dfs = {"Header": fake_data_header, "metadata": metadata}

    result = validate(fake_dfs)
    assert result == {"Header": [0, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="564",
    message="Child was missing or away from placement without authorisation and the date started is blank.",
    affected_fields=["MISSING", "MIS_START"],
    tables=["Missing"],
)
def validate(dfs):
    if "Missing" not in dfs:
        return {}
    else:
        mis = dfs["Missing"]
        error_mask = mis["MISSING"].isin(["M", "A", "m", "a"]) & mis["MIS_START"].isna()
        return {"Missing": mis.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"MISSING": "M", "MIS_START": pd.NA},  # 0  Fails
            {"MISSING": "A", "MIS_START": "07/02/2020"},  # 1
            {"MISSING": "A", "MIS_START": "03/02/2020"},  # 2
            {"MISSING": pd.NA, "MIS_START": pd.NA},  # 3
            {"MISSING": "M", "MIS_START": pd.NA},  # 4  Fails
            {"MISSING": "A", "MIS_START": "13/02/2020"},  # 5
        ]
    )

    fake_dfs = {"Missing": fake_data}

    assert validate(fake_dfs) == {"Missing": [0, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="333",
    message="Date should be placed for adoption must be on or prior to the date of matching child with adopter(s).",
    affected_fields=["DATE_INT"],
    tables=["AD1"],
)
def validate(dfs):
    if "AD1" not in dfs:
        return {}
    else:
        adt = dfs["AD1"]
        adt["DATE_MATCH"] = pd.to_datetime(
            adt["DATE_MATCH"], format="%d/%m/%Y", errors="coerce"
        )
        adt["DATE_INT"] = pd.to_datetime(
            adt["DATE_INT"], format="%d/%m/%Y", errors="coerce"
        )

        # If <DATE_MATCH> provided, then <DATE_INT> must also be provided and be <= <DATE_MATCH>
        mask1 = adt["DATE_MATCH"].notna() & adt["DATE_INT"].isna()
        mask2 = (
            adt["DATE_MATCH"].notna()
            & adt["DATE_INT"].notna()
            & (adt["DATE_INT"] > adt["DATE_MATCH"])
        )
        mask = mask1 | mask2

        return {"AD1": adt.index[mask].to_list()}


def test_validate():
    import pandas as pd

    fake_adt = pd.DataFrame(
        [
            {"DATE_INT": "01/06/2020", "DATE_MATCH": "05/06/2020"},  # 0
            {"DATE_INT": "02/06/2020", "DATE_MATCH": pd.NA},  # 1
            {"DATE_INT": "03/06/2020", "DATE_MATCH": "01/06/2020"},  # 2  Fails
            {"DATE_INT": "04/06/2020", "DATE_MATCH": "02/06/2020"},  # 3  Fails
            {"DATE_INT": pd.NA, "DATE_MATCH": "05/06/2020"},  # 4  Fails
        ]
    )

    fake_dfs = {"AD1": fake_adt}

    result = validate(fake_dfs)

    assert result == {"AD1": [2, 3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="304",
    message="Date unaccompanied asylum-seeking child (UASC) status ceased must be on or before the 18th birthday of a child.",
    affected_fields=["DUC"],
    tables=["UASC"],
)
def validate(dfs):
    if "UASC" not in dfs:
        return {}
    else:
        uasc = dfs["UASC"]
        uasc["DOB"] = pd.to_datetime(uasc["DOB"], format="%d/%m/%Y", errors="coerce")
        uasc["DUC"] = pd.to_datetime(uasc["DUC"], format="%d/%m/%Y", errors="coerce")
        mask = uasc["DUC"].notna() & (
            uasc["DUC"] > uasc["DOB"] + pd.offsets.DateOffset(years=18)
        )

        return {"UASC": uasc.index[mask].to_list()}


def test_validate():
    import pandas as pd

    fake_uasc = pd.DataFrame(
        [
            {"DOB": "01/06/2000", "DUC": "05/06/2019"},  # 0 Fails
            {"DOB": "02/06/2000", "DUC": pd.NA},  # 1
            {"DOB": "03/06/2000", "DUC": "01/06/2015"},  # 2
            {"DOB": "04/06/2000", "DUC": "02/06/2020"},  # 3 Fails
            {"DOB": pd.NA, "DUC": "05/06/2020"},  # 4
        ]
    )

    fake_dfs = {"UASC": fake_uasc}

    result = validate(fake_dfs)

    assert result == {"UASC": [0, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="584",
    message="Date of decision that the child should be placed for adoption this year is different from that recorded last year, but the decision to placed the child for adoption changed and the child should no longer be placed for adoption.",
    affected_fields=["DATE_PLACED"],
    tables=["PlacedAdoption", "PlacedAdoption_last"],
)
def validate(dfs):
    # If <CURRENT_COLLECTION_YEAR> -1 <DATE_PLACED> has been provided and
    # <DATE_PLACED_CEASED> and or <REASON_PLACED_CEASED> is not NULL then <CURRENT_COLLECTION_YEAR>
    # <DATE_PLACED> must be > <CURRENT_COLLECTION_YEAR> -1 <DATE_PLACED>

    if "PlacedAdoption" not in dfs or "PlacedAdoption_last" not in dfs:
        return {}
    else:
        # Rule logic: fail rows where date_placed_current is earlier than date _placed_last
        # in instances where there is no _reason_placed_ceased, and where there is a date_placed_last

        df = dfs["PlacedAdoption"]
        df["index"] = df.index
        pa_last = dfs["PlacedAdoption_last"]

        pa_last = pa_last[
            (pa_last["DATE_PLACED_CEASED"].notna())
            | (pa_last["REASON_PLACED_CEASED"].notna())
        ]
        df_merged = df.merge(
            pa_last, how="inner", on="CHILD", suffixes=("_current", "_last")
        )
        dp_earlier_list = list(
            df_merged[
                df_merged["DATE_PLACED_current"] <= df_merged["DATE_PLACED_last"]
            ].index
        )

        na_current_list = list(
            df_merged[df_merged["DATE_PLACED_current"].isna()]["index"]
        )

        error_mask = df["index"].isin(dp_earlier_list) | df["index"].isin(
            na_current_list
        )

        return {"PlacedAdoption": df.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": "01/02/2020",
                "REASON_PLACED_CEASED": "xx",
            },  # 0 Fail, test neither null
            {
                "CHILD": "child2",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": "01/02/2021",
                "REASON_PLACED_CEASED": "xx",
            },  # 1 Fail, test DPC null
            {
                "CHILD": "child3",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": "01/02/2020",
                "REASON_PLACED_CEASED": "xx",
            },  # 2 Fail, test RPC null
            {
                "CHILD": "child4",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": pd.NA,
                "REASON_PLACED_CEASED": pd.NA,
            },
            {
                "CHILD": "child5",
                "DATE_PLACED": "01/02/2021",
                "DATE_PLACED_CEASED": "01/02/2021",
                "REASON_PLACED_CEASED": pd.NA,
            },  # 4 pass
            {
                "CHILD": "child6",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": "01/02/2021",
                "REASON_PLACED_CEASED": "XX",
            },  # 5 pass, test DP_last null
            {
                "CHILD": "child7",
                "DATE_PLACED": pd.NA,
                "DATE_PLACED_CEASED": "01/02/2021",
                "REASON_PLACED_CEASED": "XX",
            },  # 6 fail, DP null
        ]
    )

    fake_last = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "DATE_PLACED": "01/02/2021",
                "DATE_PLACED_CEASED": "01/02/2021",
                "REASON_PLACED_CEASED": "xx",
            },  # 0
            {
                "CHILD": "child2",
                "DATE_PLACED": "01/02/2021",
                "DATE_PLACED_CEASED": pd.NA,
                "REASON_PLACED_CEASED": "xx",
            },  # 1
            {
                "CHILD": "child3",
                "DATE_PLACED": "01/02/2021",
                "DATE_PLACED_CEASED": "01/02/2021",
                "REASON_PLACED_CEASED": pd.NA,
            },  # 2
            {
                "CHILD": "child4",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": "01/02/2020",
                "REASON_PLACED_CEASED": "xx",
            },  # 3
            {
                "CHILD": "child5",
                "DATE_PLACED": "01/02/2020",
                "DATE_PLACED_CEASED": "02/02/2020",
                "REASON_PLACED_CEASED": "xx",
            },  # 4
            {
                "CHILD": "child6",
                "DATE_PLACED": pd.NA,
                "DATE_PLACED_CEASED": "01/02/2020",
                "REASON_PLACED_CEASED": "xx",
            },  # 5
            {
                "CHILD": "child7",
                "DATE_PLACED": "01/01/2020",
                "DATE_PLACED_CEASED": "01/02/2020",
                "REASON_PLACED_CEASED": "xx",
            },  # 6
        ]
    )

    fake_data["DATE_PLACED"] = pd.to_datetime(
        fake_data["DATE_PLACED"], format="%d/%m/%Y"
    )
    fake_data["DATE_PLACED_CEASED"] = pd.to_datetime(
        fake_data["DATE_PLACED_CEASED"], format="%d/%m/%Y"
    )

    fake_last["DATE_PLACED"] = pd.to_datetime(
        fake_last["DATE_PLACED"], format="%d/%m/%Y"
    )
    fake_last["DATE_PLACED_CEASED"] = pd.to_datetime(
        fake_last["DATE_PLACED_CEASED"], format="%d/%m/%Y"
    )

    fake_dfs = {"PlacedAdoption": fake_data, "PlacedAdoption_last": fake_last}

    assert validate(fake_dfs) == {"PlacedAdoption": [0, 1, 2, 3, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="582",
    message="Child is showing as ceasing to be looked after due to adoption, but the date of adoption is the same as the date reported for being matched with adopters. In most circumstances we expect the date the child is matched with adopters to be before the date of adoption.",
    affected_fields=["DEC", "DATE_MATCH"],
    tables=["Episodes", "AD1"],
)
def validate(dfs):
    # If <REC> = 'E11' or 'E12' then that episode <DEC> should not = <DATE_MATCH>
    # Apply for 2023 onwards
    if "Episodes" not in dfs or "AD1" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        ad1 = dfs["AD1"]

        episodes = episodes.reset_index()

        rec_codes = ["E11", "E12"]

        episodes = episodes[episodes["REC"].isin(rec_codes)]

        merged_df = episodes.merge(ad1, how="left", on="CHILD").set_index("index")

        episodes_with_errors = merged_df[merged_df["DEC"] == merged_df["DATE_MATCH"]]

        error_mask = episodes.index.isin(episodes_with_errors.index)

        error_locations = episodes.index[error_mask]

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        {
            "CHILD": ["0", "A", "B", "C", "D", "E", "F", "G", "H", "I"],
            "REC": ["x", "E11", "E12", "E11", "E12", "E11", "E12", "E11", "E11", "A3"],
            "DEC": [
                "01/01/2020",
                "01/01/2020",
                "01/01/2020",
                "01/01/2020",
                "01/01/2020",
                "01/01/2020",
                "01/01/2020",
                "01/01/2020",
                "01/01/2020",
                "01/01/2020",
            ],
        }
    )
    fake_data_ad1 = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C", "D", "E", "F", "G", "H"],
            "DATE_MATCH": [
                pd.NA,
                "02/01/2020",
                "02/01/2020",
                "02/01/2020",
                "02/01/2020",
                "02/01/2020",
                "02/01/2020",
                "01/01/2020",
            ],
        }
    )

    fake_dfs = {
        "Episodes": fake_data_episodes,
        "AD1": fake_data_ad1,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="551",
    message="Child has been placed for adoption but there is no date of the decision that the child should be placed for adoption.",
    affected_fields=["DATE_PLACED", "PLACE"],
    tables=["Episodes", "PlacedAdoption"],
)
def validate(dfs):
    if "Episodes" not in dfs or "PlacedAdoption" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        placedAdoptions = dfs["PlacedAdoption"]

        episodes = episodes.reset_index()

        place_codes = ["A3", "A4", "A5", "A6"]

        placeEpisodes = episodes[episodes["PLACE"].isin(place_codes)]

        merged = placeEpisodes.merge(placedAdoptions, how="left", on="CHILD").set_index(
            "index"
        )

        episodes_with_errors = merged[merged["DATE_PLACED"].isna()]

        error_mask = episodes.index.isin(episodes_with_errors.index)

        error_locations = episodes.index[error_mask]

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        {
            "CHILD": ["0", "A", "B", "C", "D", "E", "F", "G", "H", "I"],
            "PLACE": ["x", "A3", "A4", "A5", "A6", "A3", "A4", "x", "A3", "A5"],
        }
    )
    fake_data_placed_adoption = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C", "D", "E", "F", "G", "H"],
            "DATE_PLACED": [
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "01/01/2020",
                "15/04/2020",
                pd.NA,
                "28th Jan 1930",
            ],
        }
    )

    fake_dfs = {
        "Episodes": fake_data_episodes,
        "PlacedAdoption": fake_data_placed_adoption,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 2, 3, 4, 9]}



import pandas as pd

from lac_validator.fixtures import current_episodes, previous_episodes
from lac_validator.rule_engine import rule_definition
from lac_validator.rules.rule_utils import field_different_from_previous


@rule_definition(
    code="503H",
    message="The placement LA in first episode does not match open episode at end of last year.",
    affected_fields=["PL_LA"],
    tables=["Episodes", "Episodes_last"],
)
def validate(dfs):
    return field_different_from_previous(dfs, field="PL_LA")


def test_validate():
    fake_epi = current_episodes.copy()
    fake_epi_last = previous_episodes.copy()

    fake_epi["DECOM"] = pd.to_datetime(
        fake_epi["DECOM"], format="%d/%m/%Y", errors="coerce"
    )
    fake_epi_last["DECOM"] = pd.to_datetime(
        fake_epi_last["DECOM"], format="%d/%m/%Y", errors="coerce"
    )
    fake_epi_last["DEC"] = pd.to_datetime(
        fake_epi_last["DEC"], format="%d/%m/%Y", errors="coerce"
    )

    # For the test, this doesn't affect anything except how the data is handled
    # by the field_different function, so it doesn't matter that the years
    # in the test dfs are not accurate to this collection year
    # As seen in conftest collection year is read in as XXXX/XX and stored in metadata
    # as the first four characters using [:4] in _process_metadata in datastore```
    metadata = {"collectionYear": "2022"}

    fake_dfs = {
        "Episodes": fake_epi,
        "Episodes_last": fake_epi_last,
        "metadata": metadata,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [4, 6]}



from io import TextIOWrapper

import pandas as pd

from lac_validator.types import UploadedFile, UploadError


# might be worth adding 'collection_end' as an argument, though not specified in definition
def add_col_to_episodes_CONTINUOUSLY_LOOKED_AFTER(
    episodes, collection_start, collection_end
):
    """
    adds a True/False column called 'CONTINUOUSLY_LOOKED_AFTER' to the 'Episodes' table,
    which is True if they appear to have been looked after since `collection_start`
    (NB: currently modifies, and returns reference to, the original DataFrame that's passed in)

    implemented as defined in:
    https://assets.publishing.service.gov.uk/government/uploads/system/uploads/attachment_data/file/1023549/CLA_SSDA903_2021-22_Validation_rules_Version_1.2.pdf

    ie. a child is continuously looked after unless they have any episodes which either:
        - end after collection_start with LS is V3/V4
        - end after after collection_start where REC isn't X1
        - begin after collection_start where RNE is S
    """
    eps_copy = episodes.copy()

    # Set datetime types
    collection_start_dt = pd.to_datetime(
        collection_start, format="%d/%m/%Y", errors="coerce"
    )
    collection_end_dt = pd.to_datetime(
        collection_end, format="%d/%m/%Y", errors="coerce"
    )
    eps_copy["DEC_dt"] = pd.to_datetime(
        eps_copy["DEC"], format="%d/%m/%Y", errors="coerce"
    )
    eps_copy["DECOM_dt"] = pd.to_datetime(
        eps_copy["DECOM"], format="%d/%m/%Y", errors="coerce"
    )

    eps_copy["during_year"] = (eps_copy["DECOM_dt"] <= collection_end_dt) & (
        (eps_copy["DEC_dt"] >= collection_start_dt) | eps_copy["DEC_dt"].isna()
    )

    # Find CHILD id's displaying the telltale signs of <LOOKED_AFTER_CONTINUOUSLY> = 'N'
    enters_care = (
        (eps_copy["RNE"] == "S")
        & (eps_copy["DECOM_dt"] > collection_start_dt)
        & (eps_copy["DECOM_dt"] <= collection_end_dt)
    )
    leaves_care = (
        (eps_copy["REC"] != "X1")
        & (eps_copy["DEC_dt"] >= collection_start_dt)
        & (eps_copy["DEC_dt"] < collection_end_dt)
    )
    V3_V4_episode = eps_copy["LS"].isin(["V3", "V4"]) & eps_copy["during_year"]

    # if any episode fits one of these, the child was not continuously looked after
    eps_copy["implies_not_continuous"] = enters_care | leaves_care | V3_V4_episode

    in_care_at_some_point = eps_copy.groupby("CHILD")["during_year"].transform("max")
    not_in_care_at_some_point = eps_copy.groupby("CHILD")[
        "implies_not_continuous"
    ].transform("max")

    # the "== True/False" is to cope with weird situations which cause these to potentially be float dtype,
    # which i don't have time to get into right now, but may be to do with missing DECOMs
    episodes["CONTINUOUSLY_LOOKED_AFTER"] = (in_care_at_some_point == True) & (
        not_in_care_at_some_point == False
    )

    return episodes


def add_col_to_tables_CONTINUOUSLY_LOOKED_AFTER(dfs, required_tables=None):
    """
    Takes: the `dfs` dict and `required_tables` (table name or list of table names)
    Returns: The Dataframe or list of Dataframes named in `required_tables`,
             now with an additional column 'CONTINUOUSLY_LOOKED_AFTER'

    (NB: currently modifies, and returns references to, the original DataFrames in `dfs`)
    """
    CLA_col = "CONTINUOUSLY_LOOKED_AFTER"
    try:
        metadata = dfs["metadata"]
        eps = dfs["Episodes"]
    except KeyError:
        raise KeyError(
            f"Ensure 'Episodes' and 'metadata' in `dfs` before attempting to determine "
            + f"CONTINUOUSLY_LOOKED_AFTER -- only received: {', '.join(str(i) for i in dfs.keys())}"
        )

    eps_with_CLA = add_col_to_episodes_CONTINUOUSLY_LOOKED_AFTER(
        eps, metadata["collection_start"], metadata["collection_end"]
    )
    known_CLA = eps_with_CLA.drop_duplicates("CHILD")

    if isinstance(required_tables, str):
        required_tables = [
            required_tables,
        ]
        return_df_not_list = True
    else:
        return_df_not_list = False

    if len(set(required_tables) - set(dfs.keys())) > 0:
        raise KeyError(
            "required_tables should be a key, or a collection of keys, of the dfs dict"
        )
    elif "metadata" in required_tables:
        raise ValueError("cannot add CONTINUOUSLY_LOOKED_AFTER to 'metadata'")

    else:
        out = []
        for table_name in required_tables:
            if table_name == "Episodes":
                df = eps_with_CLA
            else:
                df = dfs[table_name]
                merged = pd.merge(
                    left=df["CHILD"], right=known_CLA, on="CHILD", how="left"
                )
                df[CLA_col] = merged[CLA_col]
                df[CLA_col].fillna(False, inplace=True)
            out.append(df)
    if return_df_not_list and len(out) == 1:
        return out[0]
    else:
        return out


# list[UploadedFile | PyodideFile | str] is syntax that is available only in python 3.10
# TODO update to python3.10 there is no particular reason why this project needs to be 3.9


def process_uploaded_files(input_files: dict[str, list]) -> list[UploadedFile]:
    """
    :param dict frontend_file_dict: dict of lists showing file content and the field into which user uploaded file.

    :return: files of UploadedFile type as ingress.read_from_text expects.
    """
    # children's homes field: "description" == "CH lookup"
    # social care providers field: "description" == "SCP lookup"
    # current year: description="This year"
    # Previous year: description="Prev year"

    uploaded_files = []
    for field_name, file_refs in input_files.items():
        for file_ref in file_refs:
            # name attribute depends on whether on file type. CLI and API send different types.
            try:
                file_text = file_ref.read()
                filename = file_ref.filename
            except AttributeError:
                if isinstance(file_ref, TextIOWrapper):
                    # Text IO wrapper object from command line
                    filename = file_ref.name
                elif isinstance(file_ref, str):
                    # string path
                    filename = file_ref
                else:
                    raise UploadError("file format is not recognised.")

                with open(filename, "rb") as f:
                    file_text = f.read()

            uploaded_files.append(
                UploadedFile(
                    name=filename, description=field_name, file_content=file_text
                )
            )
    return uploaded_files



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="226",
    message="Reason for placement change is not required.",
    affected_fields=["REASON_PLACE_CHANGE", "PLACE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]

        code_list = ["T0", "T1", "T2", "T3", "T4"]

        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        # create column to see previous REASON_PLACE_CHANGE
        episodes = episodes.sort_values(["CHILD", "DECOM"])
        episodes["PREVIOUS_REASON"] = episodes.groupby("CHILD")[
            "REASON_PLACE_CHANGE"
        ].shift(1)
        # If <PL> = 'T0'; 'T1'; 'T2'; 'T3' or 'T4' then <REASON_PLACE_CHANGE> should be null in current episode and current episode - 1
        mask = episodes["PLACE"].isin(code_list) & (
            episodes["REASON_PLACE_CHANGE"].notna()
            | episodes["PREVIOUS_REASON"].notna()
        )

        # error locations
        error_locs = episodes.index[mask]
    return {"Episodes": error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_data_epi = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/01/2020",
                "PLACE": "U1",
                "REASON_PLACE_CHANGE": "XXX",
            },  # 0
            {
                "CHILD": "111",
                "DECOM": "11/01/2020",
                "PLACE": "T1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 1 Fail
            {
                "CHILD": "111",
                "DECOM": "22/01/2020",
                "PLACE": "X1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 2
            {
                "CHILD": "123",
                "DECOM": "01/01/2020",
                "PLACE": "U1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 3
            {
                "CHILD": "123",
                "DECOM": "11/01/2020",
                "PLACE": "X1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 4
            {
                "CHILD": "333",
                "DECOM": "01/01/2020",
                "PLACE": "T2",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 5
            {
                "CHILD": "333",
                "DECOM": "22/01/2020",
                "PLACE": "X1",
                "REASON_PLACE_CHANGE": "CHANGE",
            },  # 6
            {
                "CHILD": "333",
                "DECOM": "11/01/2020",
                "PLACE": "U1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 7
            {
                "CHILD": "444",
                "DECOM": "22/01/2020",
                "PLACE": "T1",
                "REASON_PLACE_CHANGE": "XXX",
            },  # 8 Fail
            {
                "CHILD": "444",
                "DECOM": "11/01/2020",
                "PLACE": "X1",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 9
            {
                "CHILD": "444",
                "DECOM": "01/01/2020",
                "PLACE": "T3",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 10 Pass
            {
                "CHILD": "666",
                "DECOM": "01/01/2020",
                "PLACE": "T4",
                "REASON_PLACE_CHANGE": pd.NA,
            },  # 11 Pass
        ]
    )
    fake_dfs = {"Episodes": fake_data_epi}

    result = validate(fake_dfs)
    assert result == {"Episodes": [1, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="151",
    message="All data items relating to a childs adoption must be coded or left blank.",
    affected_fields=[
        "DATE_INT",
        "DATE_MATCH",
        "FOSTER_CARE",
        "NB_ADOPTER",
        "SEX_ADOPTR",
        "LS_ADOPTR",
    ],
    tables=["AD1"],
)
def validate(dfs):
    if "AD1" not in dfs:
        return {}
    else:
        ad1 = dfs["AD1"]
        na_date_int = ad1["DATE_INT"].isna()
        na_date_match = ad1["DATE_MATCH"].isna()
        na_foster_care = ad1["FOSTER_CARE"].isna()
        na_nb_adoptr = ad1["NB_ADOPTR"].isna()
        na_sex_adoptr = ad1["SEX_ADOPTR"].isna()
        na_lsadoptr = ad1["LS_ADOPTR"].isna()

        ad1_not_null = (
            ~na_date_int
            & ~na_date_match
            & ~na_foster_care
            & ~na_nb_adoptr
            & ~na_sex_adoptr
            & ~na_lsadoptr
        )

        validation_error = (
            ~na_date_int
            | ~na_date_match
            | ~na_foster_care
            | ~na_nb_adoptr
            | ~na_sex_adoptr
            | ~na_lsadoptr
        ) & ~ad1_not_null
        validation_error_locations = ad1.index[validation_error]

        return {"AD1": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DATE_INT": [
                pd.NA,
                "01/01/2021",
                "01/01/2021",
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "01/01/2021",
            ],
            "DATE_MATCH": [
                pd.NA,
                "01/01/2021",
                pd.NA,
                "01/01/2021",
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
            ],
            "FOSTER_CARE": [
                pd.NA,
                "01/01/2021",
                pd.NA,
                pd.NA,
                "01/01/2021",
                pd.NA,
                pd.NA,
                pd.NA,
                "01/01/2021",
            ],
            "NB_ADOPTR": [
                pd.NA,
                "01/01/2021",
                pd.NA,
                pd.NA,
                pd.NA,
                "01/01/2021",
                pd.NA,
                pd.NA,
                pd.NA,
            ],
            "SEX_ADOPTR": [
                pd.NA,
                "01/01/2021",
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "01/01/2021",
                pd.NA,
                "01/01/2021",
            ],
            "LS_ADOPTR": [
                pd.NA,
                "01/01/2021",
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                pd.NA,
                "01/01/2021",
                pd.NA,
            ],
        }
    )

    fake_dfs = {"AD1": fake_data}

    result = validate(fake_dfs)
    assert result == {"AD1": [2, 3, 4, 5, 6, 7, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="559",
    message="Date of decision that a child should be placed for adoption was not in the current year but the date of the decision that the child should be placed for adoption was not completed in a previous return.",
    affected_fields=["DATE_PLACED"],
    tables=["PlacedAdoption", "PlacedAdoption_last"],
)
def validate(dfs):
    if "PlacedAdoption" not in dfs or "PlacedAdoption_last" not in dfs:
        return {}
    else:
        placed_adoption = dfs["PlacedAdoption"]
        pa_last = dfs["PlacedAdoption_last"]
        collection_start = dfs["metadata"]["collection_start"]

        # convert dates to appropriate format
        pa_last["DATE_PLACED"] = pd.to_datetime(
            pa_last["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        )
        placed_adoption["DATE_PLACED"] = pd.to_datetime(
            placed_adoption["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        )
        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )

        # prepare to merge
        placed_adoption.reset_index(inplace=True)
        pa_last.reset_index(inplace=True)
        merged = placed_adoption.merge(
            pa_last, how="left", on="CHILD", suffixes=["_now", "_last"]
        )

        # If <DATE_PLACED> < <COLLECTION_START_DATE> then <CURRENT_COLLECTION_YEAR> -1 <DATE_PLACED> cannot be Null
        mask = (merged["DATE_PLACED_now"] < collection_start) & merged[
            "DATE_PLACED_last"
        ].isna()
        # error locations
        error_locs = merged.loc[mask, "index_now"]
        return {"PlacedAdoption": error_locs.tolist()}


def test_validate():
    import pandas as pd

    metadata = {"collection_start": "01/04/2020", "collection_end": "31/03/2021"}
    fake_placed_adoption = pd.DataFrame(
        [
            {"CHILD": 101, "DATE_PLACED": "26/05/2022"},  # 0
            {"CHILD": 102, "DATE_PLACED": "26/05/2021"},  # 1
            {"CHILD": 103, "DATE_PLACED": pd.NA},  # 2
            {"CHILD": 104, "DATE_PLACED": "01/02/2016"},  # 3
            {"CHILD": 105, "DATE_PLACED": "26/05/2019"},  # 4
        ]
    )
    fake_last_pa = pd.DataFrame(
        [
            {"CHILD": 101, "DATE_PLACED": "26/05/2000"},  # 0
            {"CHILD": 102, "DATE_PLACED": pd.NA},  # 1
            {"CHILD": 103, "DATE_PLACED": "26/05/2000"},  # 2
            {"CHILD": 104, "DATE_PLACED": "01/02/2016"},  # 3
            {"CHILD": 105, "DATE_PLACED": pd.NA},  # 4
        ]
    )
    fake_dfs = {
        "PlacedAdoption": fake_placed_adoption,
        "PlacedAdoption_last": fake_last_pa,
        "metadata": metadata,
    }

    result = validate(fake_dfs)
    assert result == {
        "PlacedAdoption": [
            4,
        ]
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1012",
    message="No other data should be returned for OC3 children who had no episodes in the current year",
    affected_fields=["CHILD"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}

    epi = dfs["Episodes"]

    error_dict = {}
    for table in ["PlacedAdoption", "Missing", "Reviews", "AD1", "PrevPerm", "OC2"]:
        if table in dfs.keys():
            df = dfs[table]
            error_dict[table] = (
                df.reset_index()
                .merge(epi, how="left", on="CHILD", indicator=True)
                .query("_merge == 'left_only'")["index"]
                .unique()
                .tolist()
            )
    return error_dict


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "555", "555", "666", "666"],
        }
    )
    fake_ado = pd.DataFrame(
        {
            "CHILD": ["777", "222", "333", "444", "555", "666"],
        }
    )  # err at 0
    fake_mis = pd.DataFrame(
        {
            "CHILD": ["111", "888", "333", "444", "555", "666"],
        }
    )  # err at 1
    fake_rev = pd.DataFrame(
        {
            "CHILD": ["111", "222", "999", "444", "555", "666"],
        }
    )  # err at 2
    fake_ad1 = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "1010", "555", "666"],
        }
    )  # err at 3
    fake_pre = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "1111", "666"],
        }
    )  # err at 4
    fake_oc2 = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "555", "1212"],
        }
    )  # err at 5

    fake_dfs = {
        "Episodes": fake_epi,
        "PlacedAdoption": fake_ado,
        "Missing": fake_mis,
        "Reviews": fake_rev,
        "AD1": fake_ad1,
        "PrevPerm": fake_pre,
        "OC2": fake_oc2,
    }

    fake_dfs_partial = {
        "Episodes": fake_epi,
        "AD1": fake_ad1,
        "PrevPerm": fake_pre,
        "OC2": fake_oc2,
    }

    assert validate(fake_dfs) == {
        "PlacedAdoption": [0],
        "Missing": [1],
        "Reviews": [2],
        "AD1": [3],
        "PrevPerm": [4],
        "OC2": [5],
    }

    assert validate(fake_dfs_partial) == {"AD1": [3], "PrevPerm": [4], "OC2": [5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="607",
    message="Child ceased to be looked after in the year, but mother field has not been completed.",
    affected_fields=["DEC", "REC", "MOTHER", "LS", "SEX"],
    tables=["Episodes", "Header"],
)
def validate(dfs):
    if "Header" not in dfs or "Episodes" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        episodes = dfs["Episodes"]
        collection_start = dfs["metadata"]["collection_start"]
        collection_end = dfs["metadata"]["collection_end"]
        code_list = ["V3", "V4"]

        # convert to datetiime format
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )

        # prepare to merge
        episodes.reset_index(inplace=True)
        header.reset_index(inplace=True)

        merged = episodes.merge(
            header, on="CHILD", how="left", suffixes=["_eps", "_er"]
        )

        # CEASED_TO_BE_LOOKED_AFTER = DEC is not null and REC is filled but not equal to X1
        CEASED_TO_BE_LOOKED_AFTER = merged["DEC"].notna() & (
            (merged["REC"] != "X1") & merged["REC"].notna()
        )
        # and <LS> not = ‘V3’ or ‘V4’
        check_LS = ~(merged["LS"].isin(code_list))
        # and <DEC> is in <CURRENT_COLLECTION_YEAR
        check_DEC = (collection_start <= merged["DEC"]) & (
            merged["DEC"] <= collection_end
        )
        # Where <CEASED_TO_BE_LOOKED_AFTER> = ‘Y’, and <LS> not = ‘V3’ or ‘V4’ and <DEC> is in <CURRENT_COLLECTION_YEAR> and <SEX> = ‘2’ then <MOTHER> should be provided.
        mask = (
            CEASED_TO_BE_LOOKED_AFTER
            & check_LS
            & check_DEC
            & (merged["SEX"] == "2")
            & (merged["MOTHER"].isna())
        )
        header_error_locs = merged.loc[mask, "index_er"]
        eps_error_locs = merged.loc[mask, "index_eps"]
        return {
            "Episodes": eps_error_locs.tolist(),
            "Header": header_error_locs.unique().tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DEC": "01/01/2021",
                "REC": pd.NA,
                "LS": "L1",
            },  # 0 Ignore: REC is null
            {
                "CHILD": 102,
                "DEC": "01/01/2001",
                "REC": "A3",
                "LS": "V4",
            },  # 1 Ignore: LS is V4
            {
                "CHILD": 103,
                "DEC": "20/12/2001",
                "REC": "E15",
                "LS": "V3",
            },  # 2 Ignore: LS is V3
            {
                "CHILD": 104,
                "DEC": "03/01/2020",
                "REC": "E46",
                "LS": "L4",
            },  # 3 Ignore: DEC out of range
            {
                "CHILD": 105,
                "DEC": "03/04/2020",
                "REC": "E48",
                "LS": "L1",
            },  # 4 FAIL
            {
                "CHILD": 105,
                "DEC": "01/07/2020",
                "REC": "X2",
                "LS": "XO",
            },  # 5 FAIL
            {
                "CHILD": 105,
                "DEC": "10/01/2021",
                "REC": "E11",
                "LS": "L4",
            },  # 6 FAIL
            {
                "CHILD": 108,
                "DEC": "11/02/2021",
                "REC": "E48",
                "LS": "L1",
            },  # 7
            {
                "CHILD": 109,
                "DEC": "25/01/2002",
                "REC": "X1",
                "LS": "XO",
            },  # 8 Ignore: REC is X1
            {
                "CHILD": 110,
                "DEC": "25/01/2021",
                "REC": "E47",
                "LS": pd.NA,
            },  # 9
            {
                "CHILD": 111,
                "DEC": pd.NA,
                "REC": "E45",
                "LS": "L4",
            },  # 10 Ignore : DEC is null
            {
                "CHILD": 112,
                "DEC": "25/08/2020",
                "REC": "E46",
                "LS": "XO",
            },  # 11
        ]
    )
    fake_data_header = pd.DataFrame(
        [
            {"CHILD": 101, "SEX": "2", "MOTHER": pd.NA},  # 0 FAIL Ignore: REC is null
            {"CHILD": 102, "SEX": "2", "MOTHER": "0"},  # 1 Ignore: LS is V4
            {"CHILD": 103, "SEX": "2", "MOTHER": pd.NA},  # 2 Ignore: LS is V3
            {"CHILD": 104, "SEX": "2", "MOTHER": pd.NA},  # 3 Ignore: DEC out of range
            {"CHILD": 105, "SEX": "2", "MOTHER": pd.NA},  # 4 FAIL
            {"CHILD": 108, "SEX": "1", "MOTHER": pd.NA},  # 5
            {"CHILD": 109, "SEX": "2", "MOTHER": "0"},  # 6 Ignore: REC is X1
            {"CHILD": 110, "SEX": "2", "MOTHER": pd.NA},  # 7 FAIL
            {"CHILD": 111, "SEX": "2", "MOTHER": "0"},  # 8 Ignore : DEC is null
            {"CHILD": 112, "SEX": "2", "MOTHER": "1"},  # 9
        ]
    )
    metadata = {"collection_start": "01/04/2020", "collection_end": "31/03/2021"}
    fake_dfs = {
        "Header": fake_data_header,
        "Episodes": fake_data_eps,
        "metadata": metadata,
    }

    result = validate(fake_dfs)
    assert result == {"Header": [4, 7], "Episodes": [4, 5, 6, 9]}



from lac_validator.fixtures import fake_INT_file, fake_INT_header
from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="INT32",
    message="Internal Check: Child should only exist once in Header.",
    affected_fields=["CHILD"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        file = dfs["Header"]

        file["index_file"] = file.index

        file["CHILD_COUNT"] = file.groupby("CHILD")["CHILD"].transform("count")

        mask = file["CHILD_COUNT"] > 1
        eps_error_locations = file.loc[mask, "index_file"]
        return {"Header": eps_error_locations.unique().tolist()}


def test_validate():
    fake_dfs = {"Header": fake_INT_header.copy()}
    result = validate(fake_dfs)
    assert result == {"Header": [0, 1, 2, 3, 4, 5, 6, 7]}
